"""Module for items related to data, such as raw csv's and data objects."""
from .dataset import *
from .load import *
from .lookup import *
from .tabular_data.acs import *
from .tabular_data.admissions import *
from .tabular_data.adult import *
from .tabular_data.compas import *
from .tabular_data.credit import *
from .tabular_data.crime import *
from .tabular_data.german import *
from .tabular_data.health import *
from .tabular_data.law import *
from .tabular_data.lipton import *
from .tabular_data.non_binary_toy import *
from .tabular_data.nursery import *
from .tabular_data.sqf import *
from .tabular_data.synthetic import *
from .tabular_data.toy import *
from .util import *



"""Load Data from .csv files."""
from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import pandas as pd

from ethicml.utility import DataTuple

from .dataset import Dataset, LegacyDataset

__all__ = ["load_data", "create_data_obj"]


def load_data(dataset: Dataset) -> DataTuple:
    """Load dataset from its CSV file.

    This function only exists for backwards compatibility. Use dataset.load() instead.

    :param dataset: dataset object
    :returns: DataTuple with dataframes of features, labels and sensitive attributes
    """
    return dataset.load()


def create_data_obj(
    filepath: Path, s_column: str, y_column: str, additional_to_drop: list[str] | None = None
) -> ConfigurableDataset:
    """Create a `ConfigurableDataset` from the given file.

    :param filepath: path to a CSV file
    :param s_column: column that represents sensitive attributes
    :param y_column: column that contains lables
    :param additional_to_drop: other columns that should be dropped (Default: None)
    :returns: Dataset object
    """
    return ConfigurableDataset(
        filepath_=filepath,
        s_column=s_column,
        y_column=y_column,
        additional_to_drop=additional_to_drop,
    )


@dataclass
class ConfigurableDataset(LegacyDataset):
    """A configurable dataset class."""

    filepath_: Optional[Path] = None
    s_column: Optional[str] = None
    y_column: Optional[str] = None
    additional_to_drop: Optional[List[str]] = None

    def __post_init__(self) -> None:
        assert (filepath := self.filepath_) is not None
        assert (s_column := self.s_column) is not None
        assert (y_column := self.y_column) is not None
        if self.additional_to_drop is None:
            self.additional_to_drop = []

        dataframe: pd.DataFrame = pd.read_csv(filepath)

        columns: list[str] = [str(x) for x in dataframe.columns.to_numpy().tolist()]
        columns.remove(s_column)
        columns.remove(y_column)
        for additional in self.additional_to_drop:
            columns.remove(additional)

        super().__init__(
            name=filepath.name,
            num_samples=len(dataframe),
            features=columns,
            cont_features=[],
            sens_attr_spec=s_column,
            class_label_spec=y_column,
            filename_or_path=filepath,
        )



"""Useful functions used in implementations."""
from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING

from ethicml.utility import DataTuple, SubgroupTuple, TestTuple

if TYPE_CHECKING:
    from ethicml.models.inprocess.in_subprocess import InAlgoRunArgs
    from ethicml.models.preprocess.pre_subprocess import PreAlgoRunArgs


def load_data_from_flags(args: PreAlgoRunArgs | InAlgoRunArgs) -> tuple[DataTuple, SubgroupTuple]:
    """Load data from the paths specified in the flags."""
    return DataTuple.from_file(Path(args["train"])), SubgroupTuple.from_file(Path(args["test"]))


def save_transformations(
    transforms: tuple[DataTuple, TestTuple], pre_algo_run_args: PreAlgoRunArgs
) -> None:
    """Save the data to the file that was specified in the commandline arguments."""
    train, test = transforms
    assert isinstance(train, DataTuple)
    assert isinstance(test, (DataTuple, SubgroupTuple))
    train.save_to_file(Path(pre_algo_run_args["new_train"]))
    test.save_to_file(Path(pre_algo_run_args["new_test"]))



"""Demographic Parity Label flipping approach."""
from __future__ import annotations
from dataclasses import dataclass
from typing_extensions import override

import numpy as np
import pandas as pd

from ethicml.utility import DataTuple, Prediction, TestTuple

from .post_algorithm import PostAlgorithm

__all__ = ["DPFlip"]


@dataclass
class DPFlip(PostAlgorithm):
    """Randomly flip a number of decisions such that perfect demographic parity is achieved."""

    @property
    @override
    def name(self) -> str:
        return "DemPar. Post Process"

    @override
    def fit(self, train_predictions: Prediction, train: DataTuple) -> DPFlip:
        return self

    @override
    def predict(self, test_predictions: Prediction, test: TestTuple, seed: int = 888) -> Prediction:
        x, y = self._fit(test, test_predictions)
        _test_preds = self._flip(
            test_predictions, test, flip_0_to_1=True, num_to_flip=x, s_group=0, seed=seed
        )
        return self._flip(_test_preds, test, flip_0_to_1=False, num_to_flip=y, s_group=1, seed=seed)

    @override
    def run(
        self,
        train_predictions: Prediction,
        train: DataTuple,
        test_predictions: Prediction,
        test: TestTuple,
        seed: int = 888,
    ) -> Prediction:
        x, y = self._fit(test, test_predictions)
        _test_preds = self._flip(
            test_predictions, test, flip_0_to_1=True, num_to_flip=x, s_group=0, seed=seed
        )
        return self._flip(_test_preds, test, flip_0_to_1=False, num_to_flip=y, s_group=1, seed=seed)

    @staticmethod
    def _flip(
        preds: Prediction,
        dt: TestTuple,
        flip_0_to_1: bool,
        num_to_flip: int,
        s_group: int,
        seed: int,
    ) -> Prediction:
        if num_to_flip >= 0:
            pre_y_val = 0 if flip_0_to_1 else 1
            post_y_val = 1 if flip_0_to_1 else 0
        else:
            pre_y_val = 1 if flip_0_to_1 else 0
            post_y_val = 0 if flip_0_to_1 else 1
            num_to_flip = abs(num_to_flip)

        _y = preds.hard[preds.hard == pre_y_val]
        _s = preds.hard[dt.s == s_group]
        idx_s_y = _y.index.intersection(_s.index)
        rng = np.random.RandomState(seed)
        idxs = list(rng.permutation(idx_s_y))
        update = pd.Series({idx: post_y_val for idx in idxs[:num_to_flip]}, dtype=preds.hard.dtype)
        preds.hard.update(update)
        return preds

    @staticmethod
    def _fit(test: TestTuple, preds: Prediction) -> tuple[int, int]:
        y_0 = preds.hard[preds.hard == 0]
        y_1 = preds.hard[preds.hard == 1]
        s_0 = test.s[test.s == 0]
        s_1 = test.s[test.s == 1]
        # Naming is nSY
        n00 = preds.hard[s_0.index.intersection(y_0.index)].count()
        n01 = preds.hard[s_0.index.intersection(y_1.index)].count()
        n10 = preds.hard[s_1.index.intersection(y_0.index)].count()
        n11 = preds.hard[s_1.index.intersection(y_1.index)].count()

        a = (((n00 + n01) * n11) - ((n10 + n11) * n01)) / (n00 + n01)
        b = (n10 + n11) / (n00 + n01)

        if b > 1:
            x = a / b
            z = 0.0
        else:
            x = 0.0
            z = a
        return int(round(x)), int(round(z))



"""Simple upsampler that makes subgroups the same size as the majority group."""
from __future__ import annotations
from dataclasses import dataclass
from enum import auto
import itertools
from typing import Optional
from typing_extensions import override

import pandas as pd
from ranzen.misc import StrEnum

from ethicml.models.inprocess.logistic_regression import LR
from ethicml.utility import DataTuple, SoftPrediction

from .pre_algorithm import PreAlgorithm, T

__all__ = ["Upsampler", "UpsampleStrategy"]


class UpsampleStrategy(StrEnum):
    """Strategy for upsampling."""

    uniform = auto()
    preferential = auto()
    naive = auto()


@dataclass
class Upsampler(PreAlgorithm):
    """Upsampler algorithm.

    Given a datatuple, create a larger datatuple such that the subgroups have a balanced number
    of samples.
    """

    strategy: UpsampleStrategy = UpsampleStrategy.uniform

    def __post_init__(self) -> None:
        self._out_size: Optional[int] = None  # pyright: ignore

    @property
    @override
    def name(self) -> str:
        return f"Upsample {self.strategy}"

    @property
    @override
    def out_size(self) -> int:
        assert self._out_size is not None
        return self._out_size

    @override
    def fit(self, train: DataTuple, seed: int = 888) -> tuple[Upsampler, DataTuple]:
        self._out_size = train.x.shape[1]
        new_train, _ = upsample(train, train, self.strategy, seed, name=self.name)
        return self, new_train

    @override
    def transform(self, data: T) -> T:
        return data.rename(f"{self.name}: {data.name}")

    @override
    def run(self, train: DataTuple, test: T, seed: int = 888) -> tuple[DataTuple, T]:
        self._out_size = train.x.shape[1]
        return upsample(train, test, self.strategy, seed, name=self.name)


def concat_datatuples(first_dt: DataTuple, second_dt: DataTuple) -> DataTuple:
    """Given 2 datatuples, concatenate them and shuffle."""
    assert (first_dt.x.columns == second_dt.x.columns).all()
    assert first_dt.s_column == second_dt.s_column
    assert first_dt.y_column == second_dt.y_column

    a_combined: pd.DataFrame = first_dt.data
    b_combined: pd.DataFrame = second_dt.data

    combined = pd.concat([a_combined, b_combined], axis="index")
    combined: pd.DataFrame = combined.sample(frac=1.0, random_state=1).reset_index(drop=True)

    return first_dt.replace_data(combined)


def upsample(
    dataset: DataTuple,
    test: T,
    strategy: UpsampleStrategy,
    seed: int,
    name: str,
) -> tuple[DataTuple, T]:
    """Upsample a datatuple.

    :param dataset: Dataset that is used to determine the imbalance.
    :param test: Test set that is upsampled along with ``dataset``.
    :param strategy: Strategy to use for upsampling.
    :param seed: Seed for the upsampling.
    :param name: Name of the upsampling strategy.
    :returns: The upsampled train and test data.
    """
    s_vals: list[int] = list(map(int, dataset.s.unique()))
    y_vals: list[int] = list(map(int, dataset.y.unique()))

    groups = itertools.product(s_vals, y_vals)

    data: dict[tuple[int, int], DataTuple] = {}
    for s, y in groups:
        s_y_mask = (dataset.s == s) & (dataset.y == y)
        data[(s, y)] = dataset.replace_data(dataset.data.loc[s_y_mask].reset_index(drop=True))

    percentages: dict[tuple[int, int], float] = {}

    vals: list[int] = []
    for key, val in data.items():
        vals.append(val.x.shape[0])

    for key, val in data.items():
        if strategy is UpsampleStrategy.naive:
            percentages[key] = max(vals) / val.x.shape[0]
        else:
            s_val: int = key[0]
            y_val: int = key[1]

            y_eq_y = dataset.y.loc[dataset.y == y_val].count()
            s_eq_s = dataset.s.loc[dataset.s == s_val].count()

            num_samples = dataset.y.count()
            num_batch = val.y.count()

            percentages[key] = round((y_eq_y * s_eq_s / (num_batch * num_samples)), 8)

    upsampled: dict[tuple[int, int], DataTuple] = {}
    all_data: pd.DataFrame
    for key, val in data.items():
        all_data = val.data.sample(
            frac=percentages[key], random_state=seed, replace=True
        ).reset_index(drop=True)
        upsampled[key] = val.replace_data(all_data)

    upsampled_datatuple: DataTuple | None = None
    for key, val in upsampled.items():
        if upsampled_datatuple is None:
            upsampled_datatuple = val
        else:
            upsampled_datatuple = concat_datatuples(upsampled_datatuple, val)
            upsampled_datatuple = upsampled_datatuple.rename(f"{name}: {dataset.name}")

    if strategy is UpsampleStrategy.preferential:
        ranker = LR()
        rank: SoftPrediction = ranker.run(dataset, dataset)

        selected: list[pd.DataFrame] = []

        all_data = dataset.data
        all_data = pd.concat(
            [all_data, pd.DataFrame(rank.soft[:, 1], columns=["preds"])], axis="columns"
        )

        for key in data:
            s_val = key[0]
            y_val = key[1]
            s_y_mask = (dataset.s == s_val) & (dataset.y == y_val)

            ascending = s_val <= 0

            if percentages[key] > 1.0:
                selected.append(all_data.loc[s_y_mask])
                percentages[key] -= 1.0

            weight = all_data.loc[s_y_mask][dataset.y_column].count()
            selected.append(
                all_data.loc[s_y_mask]
                .sort_values(by=["preds"], ascending=ascending)
                .iloc[: int(percentages[key] * weight)]
            )

        upsampled_dataframes = selected[0].drop(["preds"], axis="columns")
        for df in selected[1:]:  # iterate over the remaining selected
            upsampled_dataframes = pd.concat(
                [upsampled_dataframes, df.drop(["preds"], axis="columns")], axis="index"
            ).reset_index(drop=True)
        upsampled_datatuple = dataset.replace_data(
            upsampled_dataframes, name=f"{name}: {dataset.name}"
        )

    assert upsampled_datatuple is not None
    return upsampled_datatuple, test.rename(name=f"{name}: {test.name}")



"""Wrapper for SKLearn implementation of SVM."""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import ClassVar
from typing_extensions import override

import numpy as np
import pandas as pd
from sklearn.svm import SVC, LinearSVC

from ethicml.models.inprocess.in_algorithm import InAlgorithmDC
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["SVM"]

from ethicml.utility import KernelType


@dataclass
class SVM(InAlgorithmDC):
    """A wrapper around the SciKitLearn Support Vector Classifier (SVC) model.

    Documentation for the underlying classifier can be found
    `here <https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html>`_.

    :param C: The penalty parameter of the error term.
    :param kernel: The kernel to use.
    """

    is_fairness_algo: ClassVar[bool] = False
    C: float = field(default_factory=lambda: SVC().C)  # type: ignore[attr-defined]
    kernel: KernelType = field(
        default_factory=lambda: KernelType[SVC().kernel],  # type: ignore[attr-defined]
    )

    @property
    @override
    def name(self) -> str:
        return f"SVM ({self.kernel})"

    @override
    def fit(self, train: DataTuple, seed: int = 888) -> SVM:
        self.clf = select_svm(self.C, self.kernel, seed)
        self.clf.fit(train.x, train.y.to_numpy().ravel())
        return self

    @override
    def predict(self, test: TestTuple) -> Prediction:
        return Prediction(hard=pd.Series(self.clf.predict(test.x)))

    @override
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        clf = select_svm(self.C, self.kernel, seed)
        clf.fit(train.x, train.y.to_numpy().ravel())
        return Prediction(hard=pd.Series(clf.predict(test.x)))


def select_svm(C: float, kernel: KernelType, seed: int) -> LinearSVC | SVC:
    """Select the appropriate SVM model for the given parameters.

    :param C: The penalty parameter of the error term.
    :param kernel: The kernel to use.
    :param seed: The seed for the random number generator.
    :returns: The instantiated SVM.
    """
    random_state = np.random.RandomState(seed=seed)
    if kernel is KernelType.linear:
        return LinearSVC(C=C, dual=False, tol=1e-12, random_state=random_state)
    return SVC(C=C, kernel=kernel.value, gamma="auto", random_state=random_state)



"""The main namespace.

This namespace contains utilities and data structures.
"""
from .common import *
from .preprocessing.adjust_labels import *
from .preprocessing.biased_split import *
from .preprocessing.domain_adaptation import *
from .preprocessing.feature_binning import *
from .preprocessing.scaling import *
from .preprocessing.splits import *
from .utility.activation import *
from .utility.data_helpers import *
from .utility.data_structures import *
from .utility.heaviside import *



"""Runs given metrics on the given results."""
from __future__ import annotations
from typing import TYPE_CHECKING, Sequence

from ethicml.metrics.per_sensitive_attribute import (
    MetricNotApplicableError,
    PerSens,
    metric_per_sens,
)
from ethicml.utility.data_structures import EvalTuple, Prediction

if TYPE_CHECKING:  # the following imports are only needed for type checking
    from collections.abc import Set as AbstractSet

    from ethicml.metrics.metric import Metric


__all__ = ["run_metrics", "per_sens_metrics_check"]


def run_metrics(
    predictions: Prediction,
    actual: EvalTuple,
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
    aggregation: PerSens | AbstractSet[PerSens] = PerSens.DIFFS_RATIOS,
    use_sens_name: bool = True,
) -> dict[str, float]:
    """Run all the given metrics on the given predictions and return the results.

    :param predictions: DataFrame with predictions
    :param actual: EvalTuple with the labels
    :param metrics: list of metrics (Default: ())
    :param per_sens_metrics: list of metrics that are computed per sensitive attribute (Default: ())
    :param aggregation: Optionally specify aggregations that are performed on the per-sens metrics.
        (Default: ``DIFFS_RATIOS``)
    :param use_sens_name: if True, use the name of the senisitive variable in the returned results.
        If False, refer to the sensitive variable as "S". (Default: ``True``)
    :returns: A dictionary of all the metric results.
    """
    result: dict[str, float] = {}
    if predictions.hard.isna().any(axis=None):  # type: ignore[arg-type]
        return {"algorithm_failed": 1.0}
    for metric in metrics:
        result[metric.name] = metric.score(predictions, actual)

    for metric in per_sens_metrics:
        per_sens = metric_per_sens(predictions, actual, metric, use_sens_name)
        agg_funcs: AbstractSet[PerSens] = (
            {aggregation} if isinstance(aggregation, PerSens) else aggregation
        )
        # we can't add the aggregations directly to ``per_sens`` because then
        # we would create aggregations of aggregations
        aggregations: dict[str, float] = {}
        for agg in agg_funcs:
            aggregations.update(agg.func(per_sens))
        per_sens.update(aggregations)
        for key, value in per_sens.items():
            result[f"{metric.name}_{key}"] = value
    return result  # SUGGESTION: we could return a DataFrame here instead of a dictionary


def per_sens_metrics_check(per_sens_metrics: Sequence[Metric]) -> None:
    """Check if the given metrics allow application per sensitive attribute."""
    for metric in per_sens_metrics:
        if not metric.apply_per_sensitive:
            raise MetricNotApplicableError(
                f"Metric {metric.name} is not applicable per sensitive "
                "attribute, apply to whole dataset instead"
            )



"""Helper functions for working with DataFrames (and Series)."""
from __future__ import annotations
from typing import overload

import pandas as pd

__all__ = ["undo_one_hot", "shuffle_df"]


def shuffle_df(df: pd.DataFrame, random_state: int) -> pd.DataFrame:
    """Shuffle a given dataframe."""
    return df.sample(frac=1.0, random_state=random_state).reset_index(drop=True)


@overload
def undo_one_hot(df: pd.DataFrame, new_column_name: None = None) -> pd.Series:
    ...


@overload
def undo_one_hot(df: pd.DataFrame, new_column_name: str) -> pd.DataFrame:
    ...


def undo_one_hot(df: pd.DataFrame, new_column_name: str | None = None) -> pd.Series | pd.DataFrame:
    """Undo one-hot encoding."""
    # we have to overwrite the column names because `idxmax` uses the column names
    df.columns = pd.Index(range(df.shape[1]))
    result = df.idxmax(axis="columns")
    if new_column_name is not None:
        return result.to_frame(name=new_column_name)
    else:
        return result



"""Algorithms by Zafar et al. for Demographic Parity."""
from __future__ import annotations
from abc import abstractmethod
import json
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, ClassVar, Final, NamedTuple
from typing_extensions import override

import pandas as pd

from ethicml.preprocessing.adjust_labels import LabelBinarizer
from ethicml.utility import DataTuple, HyperParamType, Prediction, TestTuple

from .installed_model import InstalledModel

__all__ = ["ZafarAccuracy", "ZafarBaseline", "ZafarEqOdds", "ZafarEqOpp", "ZafarFairness"]


SUB_DIR_IMPACT: Final = "src.disparate_impact.run_classifier"
SUB_DIR_MISTREAT: Final = "src.disparate_mistreatment.run_classifier"


class _FitParams(NamedTuple):
    model_path: Path
    label_converter: LabelBinarizer


class _ZafarAlgorithmBase(InstalledModel):
    def __init__(self, name: str, module: str):
        super().__init__(
            name=name,
            dir_name="zafar",
            url="https://github.com/wearepal/fair-classification.git",
            top_dir="fair-classification",
            use_pdm=True,
        )
        self._module = module
        self._fit_params: _FitParams | None = None

    @staticmethod
    def _create_file_in_zafar_format(
        data: DataTuple | TestTuple, file_path: Path, label_converter: LabelBinarizer
    ) -> None:
        """Save a DataTuple as a JSON file, which is extremely inefficient but what Zafar wants.

        :param data: DataTuple to save.
        :param file_path: Path to save to.
        :param label_converter: Instance of a LabelBinarizer to convert labels to Zafar's format.
        """
        out: dict[str, Any] = {'x': data.x.to_numpy().tolist(), "sensitive": {}}
        out["sensitive"][data.s.name] = data.s.to_numpy().tolist()
        if isinstance(data, DataTuple):
            data_converted = label_converter.adjust(data)
            out["class"] = (2 * data_converted.y.to_numpy() - 1).tolist()
        else:
            out["class"] = [-1 for _ in range(data.x.shape[0])]
        with file_path.open("w") as out_file:
            json.dump(out, out_file)

    @override
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            fit_params = self._fit(train, tmp_path)
            return self._predict(test, tmp_path, fit_params)

    @override
    def fit(self, train: DataTuple, seed: int = 888) -> _ZafarAlgorithmBase:
        with TemporaryDirectory() as tmpdir:
            self._fit_params = self._fit(train, tmp_path=Path(tmpdir), model_dir=self._code_path)
        return self

    @override
    def predict(self, test: TestTuple) -> Prediction:
        assert (params := self._fit_params) is not None, "call fit() first"
        with TemporaryDirectory() as tmpdir:
            return self._predict(test, tmp_path=Path(tmpdir), fit_params=params)

    def _fit(self, train: DataTuple, tmp_path: Path, model_dir: Path | None = None) -> _FitParams:
        model_path = (model_dir.resolve() if model_dir is not None else tmp_path) / "model.npy"
        label_converter = LabelBinarizer()
        train_path = tmp_path / "train.json"
        self._create_file_in_zafar_format(train, train_path, label_converter)

        cmd = self._get_fit_cmd(str(train_path), str(model_path))
        working_dir = self._code_path.resolve()
        self.call_script(cmd, cwd=working_dir)

        return _FitParams(model_path, label_converter)

    def _predict(self, test: TestTuple, tmp_path: Path, fit_params: _FitParams) -> Prediction:
        test_path = tmp_path / "test.json"
        self._create_file_in_zafar_format(test, test_path, fit_params.label_converter)
        predictions_path = tmp_path / "predictions.json"
        cmd = self._get_predict_cmd(
            str(test_path), str(fit_params.model_path), str(predictions_path)
        )
        working_dir = self._code_path.resolve()
        self.call_script(cmd, cwd=working_dir)
        predictions = predictions_path.open().read()
        predictions = json.loads(predictions)

        predictions_correct = pd.Series([0 if x == -1 else 1 for x in predictions])
        return Prediction(hard=fit_params.label_converter.post_only_labels(predictions_correct))

    @abstractmethod
    def _get_fit_cmd(self, train_name: str, model_path: str) -> list[str]:
        pass

    def _get_predict_cmd(self, test_name: str, model_path: str, output_file: str) -> list[str]:
        return ["-m", f"{self._module}.predict", test_name, model_path, output_file]


class ZafarBaseline(_ZafarAlgorithmBase):
    """Zafar without fairness."""

    is_fairness_algo: ClassVar[bool] = False

    def __init__(self) -> None:
        super().__init__(name="ZafarBaseline", module=SUB_DIR_IMPACT)

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {}

    @override
    def _get_fit_cmd(self, train_name: str, model_path: str) -> list[str]:
        return ["-m", f"{self._module}.fit", train_name, model_path, "baseline", "0"]


class ZafarAccuracy(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, *, gamma: float = 0.5):
        super().__init__(name=f"ZafarAccuracy, γ={gamma}", module=SUB_DIR_IMPACT)
        self.gamma = gamma

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {"gamma": self.gamma}

    @override
    def _get_fit_cmd(self, train_name: str, model_path: str) -> list[str]:
        return ["-m", f"{self._module}.fit", train_name, model_path, "gamma", str(self.gamma)]


class ZafarFairness(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, *, C: float = 0.001):
        super().__init__(name=f"ZafarFairness, C={C}", module=SUB_DIR_IMPACT)
        self._c = C

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {"C": self._c}

    @override
    def _get_fit_cmd(self, train_name: str, model_path: str) -> list[str]:
        return ["-m", f"{self._module}.fit", train_name, model_path, "c", str(self._c)]


class ZafarEqOpp(_ZafarAlgorithmBase):
    """Zafar for Equality of Opportunity."""

    _mode: ClassVar[str] = "fnr"  # class level constant
    _base_name: ClassVar[str] = "ZafarEqOpp"

    def __init__(self, *, tau: float = 5.0, mu: float = 1.2, eps: float = 0.0001):
        name = f"{self._base_name}, τ={tau}, μ={mu} ε={eps}"
        super().__init__(name=name, module=SUB_DIR_MISTREAT)
        self._tau = tau
        self._mu = mu
        self._eps = eps

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {"tau": self._tau, "mu": self._mu, "eps": self._eps}

    @override
    def _get_fit_cmd(self, train_name: str, model_path: str) -> list[str]:
        return [
            "-m",
            f"{self._module}.fit",
            train_name,
            model_path,
            self._mode,
            str(self._tau),
            str(self._mu),
            str(self._eps),
        ]


class ZafarEqOdds(ZafarEqOpp):
    """Zafar for Equalised Odds."""

    _mode: ClassVar[str] = "fprfnr"
    _base_name: ClassVar[str] = "ZafarEqOdds"



"""For assessing TNR."""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric

__all__ = ["TNR"]


@dataclass
class TNR(CfmMetric):
    """True negative rate."""

    _name: ClassVar[str] = "TNR"

    @override
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, f_pos, _, _ = self._confusion_matrix(prediction=prediction, actual=actual)
        return t_neg / (t_neg + f_pos)



"""Useful methods that are used in some of the data objects."""
from __future__ import annotations
from itertools import groupby
from typing import Dict, List, Mapping, NamedTuple, Sequence
from typing_extensions import TypeAlias

import pandas as pd

__all__ = [
    "DiscFeatureGroups",
    "LabelGroup",
    "LabelSpec",
    "filter_features_by_prefixes",
    "flatten_dict",
    "from_dummies",
    "get_discrete_features",
    "group_disc_feat_indices",
    "label_spec_to_feature_list",
    "reduce_feature_group",
    "spec_from_binary_cols",
    "single_col_spec",
]

DiscFeatureGroups: TypeAlias = Dict[str, List[str]]
"""A grouping of discrete features such that the result is one-hot encoded."""


class LabelGroup(NamedTuple):
    """Definition of a group of columns that should be interpreted as a single label."""

    columns: list[str]  # list of columns that belong to the group
    multiplier: int = 1  # multiplier assigned to this group (needed when combining groups)


LabelSpec: TypeAlias = Mapping[str, LabelGroup]
"""A label specification consisting of :class:`LabelGroup` entries with names."""


def filter_features_by_prefixes(features: Sequence[str], prefixes: Sequence[str]) -> list[str]:
    """Filter the features by prefixes.

    :param features: list of features names
    :param prefixes: list of prefixes
    :returns: filtered feature names
    """
    res: list[str] = []
    for name in features:
        filtered = any(name.startswith(pref) for pref in prefixes)
        if not filtered:
            res.append(name)
    return res


def get_discrete_features(
    all_feats: list[str], feats_to_remove: list[str], cont_feats: list[str]
) -> list[str]:
    """Get a list of the discrete features in a dataset.

    :param all_feats: List of all features in the dataset.
    :param feats_to_remove: List of features that aren't used.
    :param cont_feats: List of continuous features in the dataset.
    :returns: List of features not marked as continuous or to be removed.
    """
    return [
        item
        for item in filter_features_by_prefixes(all_feats, feats_to_remove)
        if item not in cont_feats
    ]


def group_disc_feat_indices(disc_feat_names: list[str], prefix_sep: str = "_") -> list[slice]:
    """Group discrete features names according to the first segment of their name.

    Returns a list of their corresponding slices (assumes order is maintained).

    :param disc_feat_names: List of discrete feature names.
    :param prefix_sep: Separator between the prefix and the rest of the name. (Default: "_")
    :returns: List of slices.
    """

    def _first_segment(feature_name: str) -> str:
        return feature_name.split(prefix_sep)[0]

    group_iter = groupby(disc_feat_names, _first_segment)

    feature_slices: list[slice] = []
    start_idx = 0
    for _, group in group_iter:
        len_group = len(list(group))
        indices = slice(start_idx, start_idx + len_group)
        feature_slices.append(indices)
        start_idx += len_group

    return feature_slices


def flatten_dict(dictionary: Mapping[str, list[str]] | None) -> list[str]:
    """Flatten a dictionary of lists by joining all lists to one big list."""
    if dictionary is None:
        return []
    return [x for inner_list in dictionary.values() for x in inner_list]


def reduce_feature_group(
    disc_feature_groups: Mapping[str, list[str]],
    feature_group: str,
    to_keep: Sequence[str],
    remaining_feature_name: str,
) -> DiscFeatureGroups:
    """Drop all features in the given feature group except the ones in to_keep.

    :param disc_feature_groups: Dictionary of feature groups.
    :param feature_group: Name of the feature group that will be replaced by ``to_keep``.
    :param to_keep: List of features that will be kept in the feature group.
    :param remaining_feature_name: Name of the dummy feature that will be used to summarize the
        removed features.
    :returns: Modified dictionary of feature groups.
    """
    # first set the given feature group to just the value that we want to keep
    features_to_keep = list(to_keep)
    # then add a new dummy feature to the feature group. `load_data()` will create this for us
    features_to_keep.append(f"{feature_group}{remaining_feature_name}")
    new_dfgs = dict(disc_feature_groups)
    new_dfgs[feature_group] = features_to_keep
    return new_dfgs


def reduce_feature_group_mut(
    disc_feature_groups: dict[str, list[str]],
    feature_group: str,
    to_keep: Sequence[str],
    remaining_feature_name: str,
) -> list[str]:
    """Drop all features in the given feature group except the ones in to_keep.

    :param disc_feature_groups: Dictionary of feature groups.
    :param feature_group: Name of the feature group that will be replaced by ``to_keep``.
    :param to_keep: List of features that will be kept in the feature group.
    :param remaining_feature_name: Name of the dummy feature that will be used to summarize the
        removed features.
    :returns: Flattened version of the modified dictionary of feature groups.
    """
    # first set the given feature group to just the value that we want to keep
    disc_feature_groups[feature_group] = list(to_keep)
    # then add a new dummy feature to the feature group. `load_data()` will create this for us
    disc_feature_groups[feature_group].append(f"{feature_group}{remaining_feature_name}")
    # then, regenerate the list of discrete features; just like it's done in the constructor
    return flatten_dict(disc_feature_groups)


def label_spec_to_feature_list(spec: LabelSpec) -> list[str]:
    """Extract all the feature column names from a dictionary of label specifications.

    :param spec: Dictionary of label specifications.
    :returns: A flattend list of all the columns occuring in the label specs.
    """
    feature_list: list[str] = []
    for group in spec.values():
        feature_list += group.columns
    return feature_list


def spec_from_binary_cols(label_defs: Mapping[str, Sequence[str]]) -> LabelSpec:
    """Create label specs for the most common case where columns contain 0s and 1s.

    :param label_defs: Mapping of label names to column names.
    :returns: Label specifications.
    """
    multiplier = 1
    label_spec = {}
    for name, columns in label_defs.items():
        label_spec[name] = LabelGroup(list(columns), multiplier=multiplier)
        num_columns = len(columns)
        # we assume here that the columns only contain 0s and 1s
        multiplier *= num_columns if num_columns > 1 else 2
    return label_spec


def single_col_spec(col: str, feature_name: str | None = None) -> LabelSpec:
    """Create a label spec for the case where the label is defined by a single column."""
    return {col if feature_name is None else feature_name: LabelGroup([col])}


def from_dummies(data: pd.DataFrame, categorical_cols: Mapping[str, Sequence[str]]) -> pd.DataFrame:
    """Convert one-hot encoded columns into categorical columns."""
    out = data.copy()

    for col_parent, filter_col in categorical_cols.items():
        if len(filter_col) > 1:
            undummified = (
                out[list(filter_col)]
                .idxmax(axis=1)
                .apply(lambda x: x.split(f"{col_parent}_", 1)[1])
            )

            out[col_parent] = undummified
            out.drop(filter_col, axis=1, inplace=True)

    return out



"""Class to describe features of the Communities and Crime dataset."""
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LegacyDataset
from ..util import filter_features_by_prefixes, flatten_dict

__all__ = ["Crime", "CrimeSplits"]


class CrimeSplits(Enum):
    """Splits for the Crime dataset."""

    RACE_BINARY = "Race-Binary"
    CUSTOM = "Custom"


@dataclass
class Crime(LegacyDataset):
    """UCI Communities and Crime dataset."""

    split: CrimeSplits = CrimeSplits.RACE_BINARY

    Splits: ClassVar[Type[CrimeSplits]] = CrimeSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "state": [
                "state_1",
                "state_2",
                "state_4",
                "state_5",
                "state_6",
                "state_8",
                "state_9",
                "state_10",
                "state_11",
                "state_12",
                "state_13",
                "state_16",
                "state_18",
                "state_19",
                "state_20",
                "state_21",
                "state_22",
                "state_23",
                "state_24",
                "state_25",
                "state_27",
                "state_28",
                "state_29",
                "state_32",
                "state_33",
                "state_34",
                "state_35",
                "state_36",
                "state_37",
                "state_38",
                "state_39",
                "state_40",
                "state_41",
                "state_42",
                "state_44",
                "state_45",
                "state_46",
                "state_47",
                "state_48",
                "state_49",
                "state_50",
                "state_51",
                "state_53",
                "state_54",
                "state_55",
                "state_56",
            ],
            ">0.06black": [">0.06black"],
            "high_crime": ["high_crime"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "population",
            "householdsize",
            "racepctblack",
            "racePctWhite",
            "racePctAsian",
            "racePctHisp",
            "agePct12t21",
            "agePct12t29",
            "agePct16t24",
            "agePct65up",
            "numbUrban",
            "pctUrban",
            "medIncome",
            "pctWWage",
            "pctWFarmSelf",
            "pctWInvInc",
            "pctWSocSec",
            "pctWPubAsst",
            "pctWRetire",
            "medFamInc",
            "perCapInc",
            "whitePerCap",
            "blackPerCap",
            "indianPerCap",
            "AsianPerCap",
            "HispPerCap",
            "NumUnderPov",
            "PctPopUnderPov",
            "PctLess9thGrade",
            "PctNotHSGrad",
            "PctBSorMore",
            "PctUnemployed",
            "PctEmploy",
            "PctEmplManu",
            "PctEmplProfServ",
            "PctOccupManu",
            "PctOccupMgmtProf",
            "MalePctDivorce",
            "MalePctNevMarr",
            "FemalePctDiv",
            "TotalPctDiv",
            "PersPerFam",
            "PctFam2Par",
            "PctKids2Par",
            "PctYoungKids2Par",
            "PctTeen2Par",
            "PctWorkMomYoungKids",
            "PctWorkMom",
            "NumIlleg",
            "PctIlleg",
            "NumImmig",
            "PctImmigRecent",
            "PctImmigRec5",
            "PctImmigRec8",
            "PctImmigRec10",
            "PctRecentImmig",
            "PctRecImmig5",
            "PctRecImmig8",
            "PctRecImmig10",
            "PctSpeakEnglOnly",
            "PctNotSpeakEnglWell",
            "PctLargHouseFam",
            "PctLargHouseOccup",
            "PersPerOccupHous",
            "PersPerOwnOccHous",
            "PersPerRentOccHous",
            "PctPersOwnOccup",
            "PctPersDenseHous",
            "PctHousLess3BR",
            "MedNumBR",
            "HousVacant",
            "PctHousOccup",
            "PctHousOwnOcc",
            "PctVacantBoarded",
            "PctVacMore6Mos",
            "MedYrHousBuilt",
            "PctHousNoPhone",
            "PctWOFullPlumb",
            "OwnOccLowQuart",
            "OwnOccMedVal",
            "OwnOccHiQuart",
            "RentLowQ",
            "RentMedian",
            "RentHighQ",
            "MedRent",
            "MedRentPctHousInc",
            "MedOwnCostPctInc",
            "MedOwnCostPctIncNoMtg",
            "NumInShelters",
            "NumStreet",
            "PctForeignBorn",
            "PctBornSameState",
            "PctSameHouse85",
            "PctSameCity85",
            "PctSameState85",
            "LandArea",
            "PopDens",
            "PctUsePubTrans",
            "LemasPctOfficDrugUn",
            "ViolentCrimesPerPop",
        ]

        features_to_remove = ["communityname", "fold"]
        features = discrete_features + continuous_features
        features = [feature for feature in features if feature not in features_to_remove]
        if self.split is CrimeSplits.RACE_BINARY:
            sens_attr_spec = ">0.06black"
            s_prefix = [">0.06black"]
            class_label_spec = "high_crime"
            class_label_prefix = ["high_crime"]
            continuous_features = filter_features_by_prefixes(
                continuous_features,
                ["Violent", "race", "white", "black", "indian", "Asian", "Hisp", "Other"],
            )
        elif self.split is CrimeSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError
        super().__init__(
            name=f"Crime {self.split.value}",
            num_samples=1994,
            filename_or_path="crime.csv",
            features=features,
            cont_features=continuous_features,
            s_feature_groups=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_feature_groups=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_feature_groups=disc_feature_groups,
        )



"""Class to describe features of the Adult dataset.

Uses FolkTables: https://github.com/zykls/folktables
Paper: https://arxiv.org/abs/2108.04884
"""
# pylint: skip-file
# Pylint will go crazy as we're reimplementing the Dataset Init.
from __future__ import annotations
from collections.abc import Sequence
import contextlib
import os
from pathlib import Path
from typing import Generator, Literal, get_args
from typing_extensions import TypeAlias, override

import numpy as np
import pandas as pd

from ethicml.utility.data_helpers import undo_one_hot
from ethicml.utility.data_structures import DataTuple

from ..dataset import Dataset, FeatureOrder, FeatureSplit
from ..util import (
    DiscFeatureGroups,
    LabelSpec,
    filter_features_by_prefixes,
    flatten_dict,
    get_discrete_features,
    label_spec_to_feature_list,
    spec_from_binary_cols,
)

__all__ = ["AcsIncome", "AcsEmployment"]

StateList: TypeAlias = Literal[
    'AL',
    'AK',
    'AZ',
    'AR',
    'CA',
    'CO',
    'CT',
    'DE',
    'FL',
    'GA',
    'HI',
    'ID',
    'IL',
    'IN',
    'IA',
    'KS',
    'KY',
    'LA',
    'ME',
    'MD',
    'MA',
    'MI',
    'MN',
    'MS',
    'MO',
    'MT',
    'NE',
    'NV',
    'NH',
    'NJ',
    'NM',
    'NY',
    'NC',
    'ND',
    'OH',
    'OK',
    'OR',
    'PA',
    'RI',
    'SC',
    'SD',
    'TN',
    'TX',
    'UT',
    'VT',
    'VA',
    'WA',
    'WV',
    'WI',
    'WY',
    'PR',
]


@contextlib.contextmanager
def _download_dir(root: Path) -> Generator[None, None, None]:
    curdir = Path.cwd()
    os.chdir(root.expanduser().resolve())
    try:
        yield
    finally:
        os.chdir(curdir)


class _AcsBase(Dataset):
    split: str
    target: str

    def __init__(
        self,
        name: str,
        root: str | Path,
        year: str,
        horizon: int,
        states: list[StateList],
        class_label_spec: str,
        class_label_prefix: list[str],
        discrete_only: bool = False,
        invert_s: bool = False,
    ):
        if isinstance(root, str):
            root = Path(root)
        self.root = root.expanduser().resolve()
        self.root.mkdir(exist_ok=True)

        self.year = year
        self.horizon = horizon
        self.survey = "person"
        self.states = states
        self._invert_s = invert_s

        state_list: tuple[str, ...] = get_args(StateList)
        assert all(state in state_list for state in states)

        state_string = "_".join(states)
        self._name = f"{name}_{year}_{horizon}_{state_string}_{self.split}"
        self._sens_attr_spec: str | LabelSpec = ""
        self._s_prefix: list[str] = []
        self._class_label_spec: str | LabelSpec = class_label_spec
        self._class_label_prefix = class_label_prefix
        self._discrete_only = discrete_only
        self._features: list[str] = []
        self._cont_features_unfiltered: list[str] = []
        self._map_to_binary = False
        self._discrete_feature_groups: DiscFeatureGroups | None = None
        self._num_samples = 0

    @property
    def name(self) -> str:
        """Name of the dataset."""
        return self._name

    def __len__(self) -> int:
        return self._num_samples

    @property
    def sens_attrs(self) -> list[str]:
        """Get the list of sensitive attributes."""
        if isinstance(self._sens_attr_spec, str):
            return [self._sens_attr_spec]
        assert isinstance(self._sens_attr_spec, dict)
        return label_spec_to_feature_list(self._sens_attr_spec)

    @property
    def class_labels(self) -> list[str]:
        """Get the list of class labels."""
        if isinstance(self._class_label_spec, str):
            return [self._class_label_spec]
        assert isinstance(self._class_label_spec, dict)
        return label_spec_to_feature_list(self._class_label_spec)

    @property
    def features_to_remove(self) -> list[str]:
        """Features that have to be removed from x."""
        to_remove: list[str] = []
        to_remove += self._s_prefix
        to_remove += self._class_label_prefix
        if self._discrete_only:
            to_remove += self.continuous_features
        return to_remove

    @override
    def feature_split(self, order: FeatureOrder = FeatureOrder.disc_first) -> FeatureSplit:
        features_to_remove = self.features_to_remove

        return {
            "x": filter_features_by_prefixes(self._features, features_to_remove),
            "s": self.sens_attrs,
            "y": self.class_labels,
        }

    @property
    def continuous_features(self) -> list[str]:
        """List of features that are continuous."""
        return filter_features_by_prefixes(self._cont_features_unfiltered, self.features_to_remove)

    @property
    def discrete_features(self) -> list[str]:
        """List of features that are discrete."""
        return get_discrete_features(
            list(self._features), self.features_to_remove, self.continuous_features
        )

    @property
    def disc_feature_groups(self) -> DiscFeatureGroups:
        """Return Dictionary of feature groups."""
        dfgs = self._discrete_feature_groups
        assert dfgs is not None
        return {k: v for k, v in dfgs.items() if k not in self.features_to_remove}

    def _backend_load(self, dataframe: pd.DataFrame, *, labels_as_features: bool) -> DataTuple:
        # +++ BELOW HERE IS A COPY OF DATASET LOAD +++

        assert isinstance(dataframe, pd.DataFrame)

        feature_split = self.feature_split()
        if labels_as_features:
            feature_split_x = feature_split["x"] + feature_split["s"] + feature_split["y"]
        else:
            feature_split_x = feature_split["x"]
        # =========================================================================================
        # Check whether we have to generate some complementary columns for binary features.
        # This happens when we have for example several races: race-asian-pac-islander etc, but we
        # want to have a an attribute called "race_other" that summarizes them all. Now the problem
        # is that this cannot be done before this point, because only here have we actually loaded
        # the data. So, we have to do it here, with all the information we can piece together.

        disc_feature_groups = self._discrete_feature_groups
        if disc_feature_groups is not None:
            for group in disc_feature_groups.values():
                if len(group) == 1:
                    continue
                for feature in group:
                    if feature in dataframe.columns:
                        continue  # nothing to do
                    missing_feature = feature

                    existing_features = [other for other in group if other in dataframe.columns]
                    assert len(existing_features) == len(group) - 1, "at most 1 feature missing"
                    # the dummy feature is the inverse of the existing feature
                    or_combination = dataframe[existing_features[0]] == 1
                    for other in existing_features[1:]:
                        or_combination |= dataframe[other] == 1
                    inverse: pd.Series = 1 - or_combination
                    dataframe = pd.concat(
                        [dataframe, inverse.to_frame(name=missing_feature)], axis="columns"
                    )

        # =========================================================================================
        x_data = dataframe[feature_split_x]
        s_data = dataframe[feature_split["s"]]
        y_data = dataframe[feature_split["y"]]

        if self._map_to_binary:
            s_data = (s_data + 1) // 2  # map from {-1, 1} to {0, 1}
            y_data = (y_data + 1) // 2  # map from {-1, 1} to {0, 1}

        if self._invert_s:
            assert s_data.nunique().to_numpy()[0] == 2, "s must be binary"
            s_data = 1 - s_data

        # the following operations remove rows if a label group is not properly one-hot encoded
        s_data = self._one_hot_encode_and_combine(s_data, label_type="s")
        y_data = self._one_hot_encode_and_combine(y_data, label_type="y")

        return DataTuple.from_df(x=x_data, s=s_data, y=y_data, name=self.name)

    def _one_hot_encode_and_combine(
        self, attributes: pd.DataFrame, label_type: Literal["s", "y"]
    ) -> pd.Series:
        """Construct a new label according to the LabelSpecs.

        :param attributes: DataFrame containing the attributes.
        :param label_type: Type of label to construct.
        :returns: A Series object with the new combined labels.
        """
        label_mapping = self._sens_attr_spec if label_type == "s" else self._class_label_spec
        if isinstance(label_mapping, str):
            return attributes[label_mapping]

        # create a Series of zeroes with the same length as the dataframe
        combination: pd.Series = pd.Series(
            0, index=range(len(attributes)), name=",".join(label_mapping)
        )

        for name, spec in label_mapping.items():
            if len(spec.columns) > 1:  # data is one-hot encoded
                raw_values = attributes[spec.columns]
                assert (raw_values.sum(axis="columns") == 1).all(), f"{name} is not one-hot"
                values = undo_one_hot(raw_values)
            else:
                values = attributes[spec.columns[0]]
            combination += spec.multiplier * values
        return combination


class AcsIncome(_AcsBase):
    """The ACS Income Dataset from EAAMO21/NeurIPS21 - Retiring Adult."""

    def __init__(
        self,
        root: str | Path,
        year: str,
        horizon: int,
        states: list[StateList],
        split: str = "Sex",
        target_threshold: int = 50_000,
        discrete_only: bool = False,
        invert_s: bool = False,
    ):
        self.split = split
        self.target = "PINCP"
        self.target_threshold = target_threshold

        self.sens_lookup = {"Sex": "SEX", "Race": "RAC1P"}

        super().__init__(
            name="ACS_Income",
            root=root,
            class_label_spec="PINCP_1",
            class_label_prefix=["PINCP"],
            year=year,
            horizon=horizon,
            states=states,
            discrete_only=discrete_only,
            invert_s=invert_s,
        )

    @staticmethod
    def cat_lookup(key: str) -> Sequence[int]:
        """Look up categories."""
        table = {
            "COW": range(1, 9),
            'MAR': range(1, 6),
            'RELP': range(18),
            'SEX': range(1, 3),
            'RAC1P': range(1, 10),
            'PINCP': range(2),
        }

        return table[key]

    @override
    def load(
        self, labels_as_features: bool = False, order: FeatureOrder = FeatureOrder.disc_first
    ) -> DataTuple:
        from folktables import ACSDataSource, adult_filter, folktables

        datasource = ACSDataSource(
            survey_year=self.year, horizon=f'{self.horizon}-Year', survey=self.survey
        )

        with _download_dir(self.root):
            dataframe = datasource.get_data(states=self.states, download=True)

        disc_feats = [
            'COW',
            'MAR',
            'RELP',
            'SEX',
            'RAC1P',
            'PINCP',
        ]

        continuous_features = [
            'AGEP',
            'OCCP',
            'POBP',
            'SCHL',
            'WKHP',
        ]

        data_obj = folktables.BasicProblem(
            features=disc_feats + continuous_features,
            target=self.target,
            target_transform=lambda x: x > self.target_threshold,
            group=self.split,
            preprocess=adult_filter,
            postprocess=lambda x: np.nan_to_num(x, nan=-1),
        )

        dataframe = data_obj._preprocess(dataframe)
        dataframe[data_obj.target] = dataframe[data_obj.target].apply(data_obj._target_transform)

        for feat in disc_feats:
            dataframe[feat] = (
                dataframe[feat]
                .astype('int')
                .astype(pd.CategoricalDtype(categories=self.cat_lookup(feat)))
            )

        dataframe[continuous_features] = dataframe[continuous_features].astype('int')

        dataframe = pd.get_dummies(dataframe[disc_feats + continuous_features])

        dataframe = dataframe.apply(data_obj._postprocess)

        cow_cols = [col for col in dataframe.columns if col.startswith('COW')]
        mar_cols = [col for col in dataframe.columns if col.startswith('MAR')]
        relp_cols = [col for col in dataframe.columns if col.startswith('RELP')]
        sex_cols = [col for col in dataframe.columns if col.startswith('SEX')]
        rac1p_cols = [col for col in dataframe.columns if col.startswith('RAC1P')]
        pincp_cols = [col for col in dataframe.columns if col.startswith('PINCP')]
        disc_feature_groups = {
            "COW": cow_cols,
            'MAR': mar_cols,
            'RELP': relp_cols,
            'SEX': sex_cols,
            'RAC1P': rac1p_cols,
            'PINCP': pincp_cols,
        }
        discrete_features = flatten_dict(disc_feature_groups)

        # set all the attributes manually that we couldn't set in the __init__
        self._features = discrete_features + continuous_features
        self._cont_features_unfiltered = continuous_features

        self._num_samples = dataframe.shape[0]

        self._discrete_feature_groups = disc_feature_groups

        if self.split == "Sex":
            self._sens_attr_spec = f"{self.sens_lookup[self.split]}_1"
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Race":
            self._sens_attr_spec = spec_from_binary_cols(
                {self.sens_lookup[self.split]: disc_feature_groups[self.sens_lookup[self.split]]}
            )
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Sex-Race":
            self._sens_attr_spec = spec_from_binary_cols(
                {
                    f"{self.sens_lookup['Sex']}": [f"{self.sens_lookup['Sex']}_1"],
                    f"{self.sens_lookup['Race']}": disc_feature_groups[
                        f"{self.sens_lookup['Race']}"
                    ],
                }
            )
            self._s_prefix = [self.sens_lookup["Sex"], self.sens_lookup["Race"]]
        else:
            raise NotImplementedError

        dataframe = dataframe.reset_index(drop=True)

        return self._backend_load(dataframe, labels_as_features=labels_as_features)


class AcsEmployment(_AcsBase):
    """The ACS Employmment Dataset from EAAMO21/NeurIPS21 - Retiring Adult."""

    def __init__(
        self,
        root: str | Path,
        year: str,
        horizon: int,
        states: list[StateList],
        split: str = "Sex",
        discrete_only: bool = False,
        invert_s: bool = False,
    ):
        self.split = split
        self.target = "ESR"

        self.sens_lookup = {"Sex": "SEX", "Race": "RAC1P"}

        super().__init__(
            name="ACS_Employment",
            root=root,
            class_label_spec="ESR_1",
            class_label_prefix=["ESR"],
            year=year,
            horizon=horizon,
            states=states,
            discrete_only=discrete_only,
            invert_s=invert_s,
        )

    @staticmethod
    def cat_lookup(key: str) -> Sequence[int]:
        """Look up categories."""
        table = {
            'SCHL': range(1, 25),
            'MAR': range(1, 6),
            'SEX': range(1, 3),
            'DIS': range(1, 3),
            'ESP': range(1, 9),
            'MIG': range(1, 4),
            'CIT': range(1, 5),
            'MIL': range(1, 5),
            'ANC': (1, 2, 4, 8),
            'NATIVITY': range(1, 3),
            'RELP': range(18),
            'DEAR': range(1, 3),
            'DEYE': range(1, 3),
            'DREM': range(1, 3),
            'RAC1P': range(1, 10),
            'ESR': range(1, 7),
        }

        return table[key]

    @override
    def load(
        self, labels_as_features: bool = False, order: FeatureOrder = FeatureOrder.disc_first
    ) -> DataTuple:
        from folktables import ACSDataSource, folktables

        datasource = ACSDataSource(
            survey_year=self.year, horizon=f'{self.horizon}-Year', survey=self.survey
        )

        with _download_dir(self.root):
            dataframe = datasource.get_data(states=self.states, download=True)

        disc_feats = [
            'SCHL',
            'MAR',
            'RELP',
            'DIS',
            'ESP',
            'CIT',
            'MIG',
            'MIL',
            'ANC',
            'NATIVITY',
            'DEAR',
            'DEYE',
            'DREM',
            'SEX',
            'RAC1P',
            'ESR',
        ]

        continuous_features = [
            'AGEP',
        ]

        data_obj = folktables.BasicProblem(
            features=disc_feats + continuous_features,
            target=self.target,
            target_transform=lambda x: x == 1,
            group=self.split,
            preprocess=lambda x: x,
            postprocess=lambda x: np.nan_to_num(x, nan=-1),
        )

        dataframe = data_obj._preprocess(dataframe)
        dataframe[data_obj.target] = dataframe[data_obj.target].apply(data_obj._target_transform)
        dataframe = dataframe.apply(data_obj._postprocess)

        for feat in disc_feats:
            dataframe[feat] = (
                dataframe[feat]
                .astype('int')
                .astype(pd.CategoricalDtype(categories=self.cat_lookup(feat)))
            )

        dataframe[continuous_features] = dataframe[continuous_features].astype('int')

        dataframe = pd.get_dummies(dataframe[disc_feats + continuous_features])

        dataframe = dataframe.apply(data_obj._postprocess)

        schl_cols = [col for col in dataframe.columns if col.startswith('SCHL')]
        mar_cols = [col for col in dataframe.columns if col.startswith('MAR')]
        relp_cols = [col for col in dataframe.columns if col.startswith('RELP')]
        dis_cols = [col for col in dataframe.columns if col.startswith('DIS')]
        esp_cols = [col for col in dataframe.columns if col.startswith('ESP')]
        cit_cols = [col for col in dataframe.columns if col.startswith('CIT')]
        mig_cols = [col for col in dataframe.columns if col.startswith('MIG')]
        mil_cols = [col for col in dataframe.columns if col.startswith('MIL')]
        anc_cols = [col for col in dataframe.columns if col.startswith('ANC')]
        nativity_cols = [col for col in dataframe.columns if col.startswith('NATIVITY')]
        dear_cols = [col for col in dataframe.columns if col.startswith('DEAR')]
        deye_cols = [col for col in dataframe.columns if col.startswith('DEYE')]
        drem_cols = [col for col in dataframe.columns if col.startswith('DREM')]
        sex_cols = [col for col in dataframe.columns if col.startswith('SEX')]
        rac1p_cols = [col for col in dataframe.columns if col.startswith('RAC1P')]
        esr_cols = [col for col in dataframe.columns if col.startswith('ESR')]
        disc_feature_groups = {
            "SCHL": schl_cols,
            'MAR': mar_cols,
            'RELP': relp_cols,
            'DIS': dis_cols,
            'ESP': esp_cols,
            'CIT': cit_cols,
            'MIG': mig_cols,
            'MIL': mil_cols,
            'ANC': anc_cols,
            'NATIVITY': nativity_cols,
            'DEAR': dear_cols,
            'DEYE': deye_cols,
            'DREM': drem_cols,
            'SEX': sex_cols,
            'RAC1P': rac1p_cols,
            'ESR': esr_cols,
        }
        discrete_features = flatten_dict(disc_feature_groups)

        # set all the attributes manually that we couldn't set in the __init__
        self._features = discrete_features + continuous_features
        self._cont_features_unfiltered = continuous_features

        self._num_samples = dataframe.shape[0]

        self._discrete_feature_groups = disc_feature_groups

        if self.split == "Sex":
            self._sens_attr_spec = f"{self.sens_lookup[self.split]}_1"
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Race":
            self._sens_attr_spec = spec_from_binary_cols(
                {self.sens_lookup[self.split]: disc_feature_groups[self.sens_lookup[self.split]]}
            )
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Sex-Race":
            self._sens_attr_spec = spec_from_binary_cols(
                {
                    f"{self.sens_lookup['Sex']}": [f"{self.sens_lookup['Sex']}_1"],
                    f"{self.sens_lookup['Race']}": disc_feature_groups[
                        f"{self.sens_lookup['Race']}"
                    ],
                }
            )
            self._s_prefix = [self.sens_lookup["Sex"], self.sens_lookup["Race"]]
        else:
            raise NotImplementedError

        dataframe = dataframe.reset_index(drop=True)

        return self._backend_load(dataframe, labels_as_features=labels_as_features)



"""For assessing TPR."""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric

__all__ = ["TPR"]


@dataclass
class TPR(CfmMetric):
    """True positive rate."""

    _name: ClassVar[str] = "TPR"

    @override
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        _, _, f_neg, t_pos = self._confusion_matrix(prediction=prediction, actual=actual)
        return t_pos / (t_pos + f_neg)



"""Implementation of VFAE."""
from __future__ import annotations
import json
from pathlib import Path
import sys
from typing import TYPE_CHECKING

from joblib import dump, load
import pandas as pd
import torch
from torch import optim
from torch.optim import Adam
from torch.utils.data import DataLoader

from ethicml.data.dataset import CSVDataset
from ethicml.data.lookup import get_dataset_obj_by_name
from ethicml.implementations.beutel import set_seed
from ethicml.utility import DataTuple, SubgroupTuple

from .pytorch_common import CustomDataset, TestDataset
from .utils import load_data_from_flags, save_transformations
from .vfae_modules import VFAENetwork, loss_function

if TYPE_CHECKING:
    from ethicml.models.preprocess.pre_subprocess import PreAlgoArgs, T
    from ethicml.models.preprocess.vfae import VfaeArgs


def fit(train: DataTuple, flags: VfaeArgs):
    """Train the model."""
    dataset = get_dataset_obj_by_name(flags["dataset"])()
    assert isinstance(dataset, CSVDataset)

    # Set up the data
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=flags["batch_size"])

    # Build Network
    model = VFAENetwork(
        dataset,
        flags["supervised"],
        train_data.xdim,
        latent_dims=flags["latent_dims"],
        z1_enc_size=flags["z1_enc_size"],
        z2_enc_size=flags["z2_enc_size"],
        z1_dec_size=flags["z1_dec_size"],
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(flags["epochs"])):
        train_model(epoch, model, train_loader, optimizer, flags)
    return model


def transform(model: VFAENetwork, dataset: T, flags: VfaeArgs) -> T:
    """Transform the dataset."""
    data: CustomDataset | TestDataset
    if isinstance(dataset, DataTuple):
        data = CustomDataset(dataset)
        loader = DataLoader(data, batch_size=flags["batch_size"], shuffle=False)
    else:
        data = TestDataset(dataset)
        loader = DataLoader(data, batch_size=flags["batch_size"], shuffle=False)

    post_train: list[list[float]] = []
    model.eval()
    with torch.no_grad():
        for sample in loader:
            if isinstance(dataset, DataTuple):
                _x, _s, _ = sample
            else:
                _x, _s = sample
            z1_mu, _ = model.encode_z1(_x, _s)
            # z1 = model.reparameterize(z1_mu, z1_logvar)
            post_train += z1_mu.data.tolist()

    return dataset.replace(x=pd.DataFrame(post_train))


def train_and_transform(
    train: DataTuple, test: SubgroupTuple, flags: VfaeArgs
) -> tuple[DataTuple, SubgroupTuple]:
    """Train the model and transform both the train dataset and the test dataset."""
    model = fit(train, flags)

    # Transform output
    return transform(model, train, flags), transform(model, test, flags)


def train_model(
    epoch: int, model: VFAENetwork, train_loader: DataLoader, optimizer: Adam, flags: VfaeArgs
) -> None:
    """Train the model."""
    model.train()
    train_loss = 0.0
    for batch_idx, (data_x, data_s, data_y) in enumerate(train_loader):
        data_x = data_x.to("cpu")
        data_s = data_s.to("cpu")
        data_y = data_y.to("cpu")
        optimizer.zero_grad()
        z1_trip, z2_trip, z1_d_trip, x_dec, y_pred = model(data_x, data_s, data_y)

        data_trip = (data_x, data_s, data_y)

        loss_tuple = loss_function(flags, z1_trip, z2_trip, z1_d_trip, data_trip, x_dec, y_pred)
        prediction_loss, reconstruction_loss, kld_loss, mmd_loss = loss_tuple
        loss = kld_loss + reconstruction_loss + prediction_loss + mmd_loss
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % 100 == 0:
            if flags["supervised"]:
                print(
                    f"train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}"
                    f"({100. * batch_idx / len(train_loader):.0f}%)]\t"
                    f"Loss: {loss.item() / len(data_x):.6f}\t"
                    f"pred_loss: {prediction_loss.item():.6f}\t"
                    f"recon_loss: {reconstruction_loss.item():.6f}\t"
                    f"kld_loss: {kld_loss.item():.6f}\t"
                    f"mmd_loss: {flags['batch_size'] * mmd_loss.item():.6f}"
                )
            else:
                print(
                    f"train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}"
                    f"({100. * batch_idx / len(train_loader):.0f}%)]\t"
                    f"Loss: {loss.item() / len(data_x):.6f}\t"
                    f"recon_loss: {reconstruction_loss.item():.6f}\t"
                    f"mmd_loss: {flags['batch_size'] * mmd_loss.item():.6f}"
                )

    print(f"====> Epoch: {epoch} Average loss: {train_loss / len(train_loader.dataset):.4f}")


def main() -> None:
    """Run model."""
    pre_algo_args: PreAlgoArgs = json.loads(sys.argv[1])
    flags: VfaeArgs = json.loads(sys.argv[2])
    if pre_algo_args["mode"] == "run":
        set_seed(pre_algo_args["seed"])
        train, test = load_data_from_flags(pre_algo_args)
        save_transformations(train_and_transform(train, test, flags), pre_algo_args)
    elif pre_algo_args["mode"] == "fit":
        set_seed(pre_algo_args["seed"])
        train = DataTuple.from_file(Path(pre_algo_args["train"]))
        enc = fit(train, flags)
        transformed_train = transform(enc, train, flags)
        transformed_train.save_to_file(Path(pre_algo_args["new_train"]))
        dump(enc, Path(pre_algo_args["model"]))
    elif pre_algo_args["mode"] == "transform":
        model = load(Path(pre_algo_args["model"]))
        test = SubgroupTuple.from_file(Path(pre_algo_args["test"]))
        transformed_test = transform(model, test, flags)
        transformed_test.save_to_file(Path(pre_algo_args["new_test"]))


if __name__ == "__main__":
    main()



"""Class to describe features of the Adult dataset."""
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LegacyDataset
from ..util import LabelSpec, flatten_dict

__all__ = ["Nursery", "NurserySplits"]


class NurserySplits(Enum):
    """Available dataset splits for the Adult dataset."""

    FINANCE = "Finance"
    CUSTOM = "Custom"


@dataclass
class Nursery(LegacyDataset):
    """UCI Adult dataset.

    :param discrete_only: If True, continuous features are dropped. (Default: False)
    :param invert_s: If True, the (binary) ``s`` values are inverted. (Default: False)
    :param split: What to use as ``s``. (Default: "Sex")
    :param binarize_nationality: If True, nationality will be USA vs rest. (Default: False)
    :param binarize_race: If True, race will be white vs rest. (Default: False)
    """

    Splits: ClassVar[Type[NurserySplits]] = NurserySplits

    split: NurserySplits = Splits.FINANCE

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "parents": ["parents_usual", "parents_pretentious", "parents_great_pret"],
            "has_nurs": [
                "has_nurs_proper",
                "has_nurs_less_proper",
                "has_nurs_improper",
                "has_nurs_critical",
                "has_nurs_very_crit",
            ],
            "form": ["form_complete", "form_completed", "form_foster", "form_incomplete"],
            "housing": ["housing_convenient", "housing_less_conv", "housing_critical"],
            "health": ["health_not_recom", "health_priority", "health_recommended"],
            "social": ["social_non-prob", "social_slightly_prob", "social_problematic"],
            "class": [
                "class_not_recom",
                "class_recommend",
                "class_very_recom",
                "class_priority",
                "class_spec_prior",
            ],
            "finance": ["finance_conv", "finance_inconv"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = ["children"]

        sens_attr_spec: str | LabelSpec
        if self.split is NurserySplits.FINANCE:
            sens_attr_spec = "finance_conv"
            s_prefix = ["finance"]
            class_label_spec = "class_not_recom"
            class_label_prefix = ["class"]
        else:
            raise NotImplementedError

        name = f"Nursery {self.split.value}"

        super().__init__(
            name=name,
            num_samples=12960,
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            sens_attr_spec=sens_attr_spec,
            class_label_spec=class_label_spec,
            filename_or_path="nursery.csv.zip",
            s_feature_groups=s_prefix,
            class_feature_groups=class_label_prefix,
            discrete_feature_groups=disc_feature_groups,
        )



"""For assessing PPV."""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric

__all__ = ["PPV"]


@dataclass
class PPV(CfmMetric):
    """Positive predictive value."""

    _name: ClassVar[str] = "PPV"

    @override
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        _, f_pos, _, t_pos = self._confusion_matrix(prediction=prediction, actual=actual)
        return t_pos / (t_pos + f_pos)



"""Common plotting functions / datastructures."""
from __future__ import annotations
from typing import Any, Literal, NamedTuple
from typing_extensions import TypeAlias

from matplotlib import legend
from matplotlib import pyplot as plt
import pandas as pd

LegendType: TypeAlias = Literal["inside", "outside"]  # pylint: disable=invalid-name
PlotType: TypeAlias = Literal["box", "cross", "scatter", "line"]  # pylint: disable=invalid-name


class DataEntry(NamedTuple):
    """Defines one element in a plot."""

    label: str
    vals: pd.DataFrame
    do_fill: bool


class PlotDef(NamedTuple):
    """All the things needed to make a plot."""

    title: str
    entries: list[DataEntry]
    legend_pos: LegendType | None = None
    legend_yanchor: float = 1.0


def common_plotting_settings(
    plot: plt.Axes, plot_def: PlotDef, xaxis_title: str, yaxis_title: str
) -> legend.Legend | None:
    """Commonly used settings for plots.

    :param plot: a pyplot plot object
    :param plot_def: a `PlotDef` that defines properties of the plot
    :param xaxis_title: label for x-axis
    :param yaxis_title: label for y-axis
    :returns: A legend object if a legend was defined; otherwise ``None``.
    """
    plot.set_xlabel(xaxis_title)
    plot.set_ylabel(yaxis_title)
    if title := plot_def.title:
        plot.set_title(title)
    plot.grid(True)

    # get handles and labels for the legend
    handles, labels = plot.get_legend_handles_labels()
    # remove the errorbars from the legend if they are there
    handles = [(h[0] if isinstance(h, tuple) else h) for h in handles]

    if plot_def.legend_pos == "inside":
        return plot.legend(handles, labels)
    if plot_def.legend_pos == "outside":
        return plot.legend(
            handles, labels, loc="upper left", bbox_to_anchor=(1, plot_def.legend_yanchor)
        )
    return None


def errorbox(
    plot: plt.Axes,
    plot_def: PlotDef,
    xaxis: tuple[str, str],
    yaxis: tuple[str, str],
    firstcolor: int = 0,
    firstshape: int = 0,
    markersize: int = 6,
    use_cross: bool = False,
) -> legend.Legend | None:
    """Generate a figure with errorboxes that reflect the std dev of an entry.

    :param plot: a pyplot Axes object
    :param plot_def: a `PlotDef` that defines properties of the plot
    :param xaxis: a tuple of two strings
    :param yaxis: a tuple of two strings
    :param firstcolor: index of the color that should be used for the first entry (Default: 0)
    :param firstshape: index of the shape that should be used for the first entry (Default: 0)
    :param markersize: size of the markers (Default: 6)
    :param use_cross: if True, use cross instead of boxes (Default: False)
    :returns: A legend object if a legend was defined; otherwise ``None``.
    """
    # ================================ constants for plots ========================================
    colors = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd"]
    colors += ["#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"]
    pale_colors = ["#aec7e8", "#ffbb78", "#98df8a", "#ff9896", "#c5b0d5"]
    pale_colors += ["#c49c94", "#f7b6d2", "#c7c7c7", "#dbdb8d", "#9edae5"]
    shapes = ["o", "X", "D", "s", "^", "v", "<", ">", "*", "p", "P"]
    hatch_pattern = ["O", "o", "|", "-", "/", "\\", "+", "x", "*"]

    xaxis_measure, yaxis_measure = xaxis[0], yaxis[0]
    filled_counter = firstcolor

    max_xmean = 0.0
    min_xmean = 1.0

    max_ymean = 0.0
    min_ymean = 1.0

    for i, entry in enumerate(plot_def.entries):
        if entry.do_fill:
            color = colors[filled_counter]
            filled_counter += 1
        else:
            color = pale_colors[i + firstcolor - filled_counter]
        i_shp = firstshape + i

        xmean: float = entry.vals[xaxis_measure].mean()
        xstd: float = entry.vals[xaxis_measure].std()
        if xmean + (0.5 * xstd) > max_xmean:
            max_xmean = xmean + (0.5 * xstd)
        if xmean - (0.5 * xstd) < min_xmean:
            min_xmean = xmean - (0.5 * xstd)

        ymean: float = entry.vals[yaxis_measure].mean()
        ystd: float = entry.vals[yaxis_measure].std()
        if ymean + (0.5 * ystd) > max_ymean:
            max_ymean = ymean + (0.5 * ystd)
        if ymean - (0.5 * ystd) < min_ymean:
            min_ymean = ymean - (0.5 * ystd)

        if use_cross:
            plot.errorbar(
                xmean,
                ymean,
                xerr=0.5 * xstd,
                yerr=0.5 * ystd,
                marker=shapes[i_shp % len(shapes)],
                linestyle="",  # no connecting lines
                color=color,
                ecolor="#555555",
                capsize=4,
                elinewidth=1.0,
                zorder=3 + 2 * i_shp,
                label=entry.label,
                markersize=markersize,
            )
        else:
            plot.bar(
                xmean,
                ystd,
                bottom=ymean - 0.5 * ystd,
                width=xstd,
                align="center",
                color="none",
                edgecolor=color,
                linewidth=3,
                zorder=3 + 2 * i_shp,
                hatch=hatch_pattern[i % len(hatch_pattern)],
            )
            plot.plot(
                xmean,
                ymean,
                marker=shapes[i_shp % len(shapes)],
                linestyle="",  # no connecting lines
                color=color,
                label=entry.label,
                zorder=4 + 2 * i_shp,
                markersize=markersize,
            )

    x_min = min_xmean * 0.99
    x_max = max_xmean * 1.01

    y_min = min_ymean * 0.99
    y_max = max_ymean * 1.01

    plot.set_xlim(x_min, x_max)
    plot.set_ylim(y_min, y_max)
    return common_plotting_settings(plot, plot_def, xaxis[1], yaxis[1])


def scatter(
    plot: plt.Axes,
    plot_def: PlotDef,
    xaxis: tuple[str, str],
    yaxis: tuple[str, str],
    startindex: int = 0,
    markersize: int = 6,
    connect_dots: bool = False,
) -> legend.Legend | None:
    """Generate a scatter plot.

    :param plot: a pyplot plot object
    :param plot_def: a `PlotDef` that defines properties of the plot
    :param xaxis: a tuple of two strings
    :param yaxis: a tuple of two strings
    :param startindex: index of color and shape that should be used for the first entry (Default: 0)
    :param markersize: size of the markers (Default: 6)
    :param connect_dots: if True, connect the data points with a line (Default: False)
    :returns: A legend object if a legend was defined; otherwise ``None``.
    """
    shapes = ["o", "X", "D", "s", "^", "v", "<", ">", "*", "p", "P"]
    colors10 = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd"]
    colors10 += ["#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"]
    xaxis_measure, yaxis_measure = xaxis[0], yaxis[0]
    filled_counter = startindex
    for i, entry in enumerate(plot_def.entries):
        additional_params: dict[str, Any]
        if entry.do_fill or connect_dots:
            additional_params = {}
            shp_index = filled_counter
            filled_counter += 1
        else:
            additional_params = dict(markerfacecolor="none")
            shp_index = i + startindex - filled_counter
        plot.plot(
            entry.vals[xaxis_measure].to_numpy(),
            entry.vals[yaxis_measure].to_numpy(),
            marker=shapes[shp_index],
            linestyle="-" if connect_dots else "",
            label=entry.label,
            color=None if connect_dots else colors10[shp_index],
            markersize=markersize,
            **additional_params,
        )
    return common_plotting_settings(plot, plot_def, xaxis[1], yaxis[1])



"""Class to describe features of the Synthetic dataset."""
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import ClassVar, Type
from typing_extensions import override

from ethicml.data.dataset import CSVDatasetDC, LabelSpecsPair
from ethicml.data.util import DiscFeatureGroups, single_col_spec

__all__ = ["SyntheticScenarios", "SyntheticTargets", "Synthetic"]


class SyntheticScenarios(Enum):
    """Scenarios for the synthetic dataset."""

    S1 = 1
    S2 = 2
    S3 = 3
    S4 = 4


class SyntheticTargets(Enum):
    """Targets for the synthetic dataset."""

    Y1 = 1
    Y2 = 2
    Y3 = 3


@dataclass
class Synthetic(CSVDatasetDC):
    r"""Dataset with synthetic data.

    ⊥ = is independent of
    ~ = is an ancestor of in the causal model used to generate the data

    Scenario 1 = X⊥S & Y⊥S.
        - This models completely fair data.
    Scenario 2 = X_2⊥S & Y_2⊥S; X_1~S, Y_1~S & Y_3~S
        - This models data where the inputs are biased. This is propogated through to the target.
    Scenario 3 = X⊥S, Y_1⊥S, Y_2⊥S; Y_3~S
        - This models data where the target is biased.
    Scenario 4 = X_2⊥S, Y_2⊥S; X_1~S, Y_1~S, Y_3~S
        - This models data where both the input and target are directly biased.
    """

    Scenarios: ClassVar[Type[SyntheticScenarios]] = SyntheticScenarios
    Targets: ClassVar[Type[SyntheticTargets]] = SyntheticTargets

    scenario: SyntheticScenarios = SyntheticScenarios.S1
    target: SyntheticTargets = SyntheticTargets.Y3
    fair: bool = False
    num_samples: int = 1_000

    def __post_init__(self) -> None:
        assert 0 < self.num_samples <= 100_000

    @property
    @override
    def continuous_features(self) -> list[str]:
        return ["x1f", "x2f", "n1", "n2"] if self.fair else ["x1", "x2", "n1", "n2"]

    @property
    @override
    def name(self) -> str:
        return (
            f"Synthetic - Scenario {self.scenario.value}, "
            f"target {self.target.value}{' fair' if self.fair else ''}"
        )

    @override
    def get_label_specs(self) -> LabelSpecsPair:
        y = single_col_spec(f"y{self.target.value}{'f' if self.fair else ''}")
        return LabelSpecsPair(s=single_col_spec("s"), y=y)

    @override
    def get_num_samples(self) -> int:
        return self.num_samples

    @override
    def get_filename_or_path(self) -> str | Path:
        return f"synthetic_scenario_{self.scenario.value}.csv"

    @property
    @override
    def unfiltered_disc_feat_groups(self) -> DiscFeatureGroups:
        return {}



"""Abstract Base Class of all algorithms in the framework."""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TypeVar
from typing_extensions import Self

from ethicml.models.algorithm_base import Algorithm
from ethicml.utility import DataTuple, SubgroupTuple

__all__ = ["PreAlgorithm"]

T = TypeVar("T", DataTuple, SubgroupTuple)


class PreAlgorithm(Algorithm, ABC):
    """Abstract Base Class for all algorithms that do pre-processing."""

    @abstractmethod
    def fit(self, train: DataTuple, seed: int = 888) -> tuple[Self, DataTuple]:
        """Fit transformer on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of Self and the test data.
        """

    @abstractmethod
    def transform(self, data: T) -> T:
        """Generate fair features with the given data.

        :param data: Data to transform.
        :returns: Transformed data.
        """

    @abstractmethod
    def run(self, train: DataTuple, test: T, seed: int = 888) -> tuple[DataTuple, T]:
        """Generate fair features with the given data.

        :param train: Data tuple of the training data.
        :param test: Data tuple of the test data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of the transforme training data and the test data.
        """

    def run_test(self, train: DataTuple, test: T, seed: int = 888) -> tuple[DataTuple, T]:
        """Run with reduced training set so that it finishes quicker.

        :param train: Data tuple of the training data.
        :param test: Data tuple of the test data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of the transforme training data and the test data.
        """
        train_testing = train.get_n_samples()
        return self.run(train_testing, test, seed)

    @property
    @abstractmethod
    def out_size(self) -> int:
        """Return the number of features to generate."""



"""Methods that are shared among the inprocess algorithms."""
from __future__ import annotations
from typing import Union

from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC, LinearSVC

from ethicml.utility import ClassifierType, KernelType

__all__ = ["LinearModel", "settings_for_svm_lr"]


LinearModel = Union[SVC, LinearSVC, LogisticRegression]


def settings_for_svm_lr(
    classifier: ClassifierType, C: float | None, kernel: KernelType | None
) -> tuple[float, KernelType | None]:
    """If necessary get the default settings for the C and kernel parameter of SVM and LR."""
    if classifier is ClassifierType.gbt:
        return 1.0, None
    if C is None:
        if classifier is ClassifierType.lr:
            C = LogisticRegression().C  # type: ignore[attr-defined]
        elif classifier is ClassifierType.svm:
            C = SVC().C  # type: ignore[attr-defined]
        else:
            raise NotImplementedError(f'Unsupported classifier "{classifier}".')

    if kernel is None:
        kernel = (
            KernelType[SVC().kernel]  # type: ignore[attr-defined]
            if classifier is ClassifierType.svm
            else None
        )
    return C, kernel



"""Evaluator for a metric per sensitive attribute class."""
from __future__ import annotations
from enum import Enum
from typing import Callable, ClassVar, Mapping

import pandas as pd

from ethicml.utility.data_structures import EvalTuple, Prediction, SoftPrediction

from .metric import Metric

__all__ = [
    "MetricNotApplicableError",
    "PerSens",
    "aggregate_over_sens",
    "diff_per_sens",
    "max_per_sens",
    "metric_per_sens",
    "min_per_sens",
    "ratio_per_sens",
]


class MetricNotApplicableError(Exception):
    """Metric Not Applicable per sensitive attribute, apply to whole dataset instead."""


def metric_per_sens(
    prediction: Prediction, actual: EvalTuple, metric: Metric, use_sens_name: bool = True
) -> dict[str, float]:
    """Compute a metric repeatedly on subsets of the data that share a senstitive attribute."""
    if not metric.apply_per_sensitive:
        raise MetricNotApplicableError(
            f"Metric {metric.name} is not applicable per sensitive "
            "attribute, apply to whole dataset instead"
        )

    assert actual.s.shape[0] == actual.y.shape[0]
    assert prediction.hard.shape[0] == actual.y.shape[0]

    per_sensitive_attr: dict[str, float] = {}

    s_column: str = actual.s_column

    for unique_s in actual.s.unique():
        mask: pd.Series = actual.s == unique_s
        subset = actual.get_s_subset(unique_s)
        pred_y: Prediction
        if isinstance(prediction, SoftPrediction):
            pred_y = SoftPrediction(soft=prediction.soft[mask], info=prediction.info)
        else:
            pred_y = Prediction(
                hard=prediction.hard.loc[mask].reset_index(drop=True), info=prediction.info
            )
        key = (s_column if use_sens_name else "S") + "_" + str(unique_s)
        per_sensitive_attr[key] = metric.score(pred_y, subset)

    return per_sensitive_attr


def aggregate_over_sens(
    per_sens_res: Mapping[str, float],
    aggregator: Callable[[float, float], float],
    infix: str,
    prefix: str = "",
    suffix: str = "",
) -> dict[str, float]:
    """Aggregate metrics over sensitive attributes.

    :param per_sens_res: Dictionary of the results.
    :param aggregator: A callable that is used to aggregate results.
    :param infix: A string that will be displayed between the sensitive attributes in the final
        metric name.
    :param prefix: A prefix for the final metric name.
    :param suffix: A suffix for the final metric name.
    :returns: Dictionary of the aggregated results.
    """
    sens_keys = sorted(per_sens_res.keys())
    aggregated_over_sens: dict[str, float] = {}

    for i, sens_key_i in enumerate(sens_keys):
        i_value: float = per_sens_res[sens_key_i]
        for j in range(i + 1, len(sens_keys)):
            key: str = f"{prefix}{sens_key_i}{infix}{sens_keys[j]}{suffix}"
            j_value: float = per_sens_res[sens_keys[j]]

            aggregated_over_sens[key] = aggregator(i_value, j_value)

    return aggregated_over_sens


def diff_per_sens(per_sens_res: dict[str, float]) -> dict[str, float]:
    """Compute the difference in the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of differences
    """
    return aggregate_over_sens(per_sens_res, aggregator=_abs_diff, infix="-")


def _abs_diff(i_value: float, j_value: float) -> float:
    return abs(i_value - j_value)


def ratio_per_sens(per_sens_res: dict[str, float]) -> dict[str, float]:
    """Compute the ratios in the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of ratios
    """
    return aggregate_over_sens(per_sens_res, aggregator=_safe_ratio, infix="÷")


def _safe_ratio(i_value: float, j_value: float) -> float:
    min_val = min(i_value, j_value)
    max_val = max(i_value, j_value)

    return min_val / max_val if max_val != 0 else float("nan")


def min_per_sens(per_sens_res: dict[str, float]) -> dict[str, float]:
    """Compute the minimum value of the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of min values
    """
    return aggregate_over_sens(per_sens_res, aggregator=min, prefix="min(", infix=",", suffix=")")


def max_per_sens(per_sens_res: dict[str, float]) -> dict[str, float]:
    """Compute the maximum value of the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of max values
    """
    return aggregate_over_sens(per_sens_res, aggregator=max, prefix="max(", infix=",", suffix=")")


class PerSens(Enum):
    """Aggregation methods for metrics that are computed per sensitive attributes."""

    _ignore_ = ["ALL", "DIFFS_RATIOS", "MIN_MAX"]  # these are class variables and not enum members

    ALL: ClassVar[frozenset[PerSens]]
    """All aggregations."""
    DIFFS = (diff_per_sens,)
    """Differences of the per-group results."""
    DIFFS_RATIOS: ClassVar[frozenset[PerSens]]
    """Equivalent to ``{DIFFS, RATIOS}``."""
    MAX = (max_per_sens,)
    """Maximum of the per-group results."""
    MIN = (min_per_sens,)
    """Minimum of the per-group results."""
    MIN_MAX: ClassVar[frozenset[PerSens]]
    """Equivalent to ``{MIN, MAX}``."""
    RATIOS = (ratio_per_sens,)
    """Ratios of the per-group results."""

    def __init__(self, agg_func: Callable[[dict[str, float]], dict[str, float]]):
        self.func = agg_func


# Unfortunately, we have to set the ClassVars outside of the class body.
# This problem is solved in Python 3.11, but we can't rely on that yet.
PerSens.DIFFS_RATIOS = frozenset({PerSens.DIFFS, PerSens.RATIOS})
PerSens.MIN_MAX = frozenset({PerSens.MIN, PerSens.MAX})
PerSens.ALL = frozenset({PerSens.DIFFS, PerSens.RATIOS, PerSens.MIN, PerSens.MAX})



"""Class to describe features of the Toy dataset."""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from ..dataset import LabelSpecsPair, StaticCSVDataset
from ..util import DiscFeatureGroups, single_col_spec

__all__ = ["Toy"]


@dataclass
class Toy(StaticCSVDataset):
    """Dataset with toy data for testing."""

    num_samples: ClassVar[int] = 400
    csv_file: ClassVar[str] = "toy.csv"

    @property
    @override
    def name(self) -> str:
        return "Toy"

    @override
    def get_label_specs(self) -> LabelSpecsPair:
        return LabelSpecsPair(s=single_col_spec("sensitive-attr"), y=single_col_spec("decision"))

    @property
    @override
    def unfiltered_disc_feat_groups(self) -> DiscFeatureGroups:
        return {
            "disc_1": ["disc_1_a", "disc_1_b", "disc_1_c", "disc_1_d", "disc_1_e"],
            "disc_2": ["disc_2_x", "disc_2_y", "disc_2_z"],
        }

    @property
    @override
    def continuous_features(self) -> list[str]:
        return ["a1", "a2"]



"""Independence of 2 variables."""
from __future__ import annotations
from typing import Type

import numpy as np
import torch

from .density_estimation import Kde


def _joint_2(
    x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 1e-10
) -> torch.Tensor:
    x = (x - x.mean()) / x.std()
    y = (y - y.mean()) / y.std()
    data = torch.cat([x.unsqueeze(-1), y.unsqueeze(-1)], -1)
    joint_density = density(data)

    nbins = int(min(50, 5.0 / joint_density.std))
    # nbins = np.sqrt( Y.size/5 )
    x_centers = torch.linspace(-2.5, 2.5, nbins)
    y_centers = torch.linspace(-2.5, 2.5, nbins)

    xx, yy = torch.meshgrid([x_centers, y_centers])
    grid = torch.cat([xx.unsqueeze(-1), yy.unsqueeze(-1)], -1)
    h2d: torch.Tensor = joint_density.pdf(grid) + damping
    h2d /= h2d.sum()
    return h2d


def hgr(x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 1e-10) -> float:
    """An estimator of the Hirschfeld-Gebelein-Renyi maximum correlation coefficient.

    This function is using Witsenhausen’s Characterization.

    HGR(x,y) is the second highest eigenvalue of the joint density on (x,y). We compute here the
    second eigenvalue on an empirical and discretized density estimated from the input data.

    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param density: so far only kde is supported
    :param damping: a damping factor
    :returns: numerical value between 0 and 1 (0: independent, 1:linked by a deterministic equation)
    """
    h2d = _joint_2(x, y, density, damping=damping)
    marginal_x = h2d.sum(dim=1).unsqueeze(1)
    marginal_y = h2d.sum(dim=0).unsqueeze(0)
    Q = h2d / (torch.sqrt(marginal_x) * torch.sqrt(marginal_y))  # noqa: N806
    return torch.svd(Q)[1][1]


def chi_2(x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 0) -> float:
    r"""The :math:`\chi^2` divergence between the joint distribution and the product of marginals.

    This is know to be the square of an upper-bound on the Hirschfeld-Gebelein-Renyi maximum
    correlation coefficient. We compute it here on an empirical and discretized density estimated
    from the input data.

    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param density: so far only kde is supported
    :param damping: a damping factor
    :returns: numerical value between 0 and :math:`\infty` (0: independent)
    """
    h2d = _joint_2(x, y, density, damping=damping)
    marginal_x = h2d.sum(dim=1).unsqueeze(1)
    marginal_y = h2d.sum(dim=0).unsqueeze(0)
    Q = h2d / (torch.sqrt(marginal_x) * torch.sqrt(marginal_y))  # noqa: N806
    return (Q**2).sum(dim=[0, 1]) - 1.0


# Independence of conditional variables


def _joint_3(
    x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde], damping: float = 1e-10
) -> torch.Tensor:
    x = (x - x.mean()) / x.std()
    y = (y - y.mean()) / y.std()
    z = (z - z.mean()) / z.std()
    data = torch.cat([x.unsqueeze(-1), y.unsqueeze(-1), z.unsqueeze(-1)], -1)
    joint_density = density(data)  # + damping

    nbins = int(min(50, 5.0 / joint_density.std))
    x_centers = torch.linspace(-2.5, 2.5, nbins)
    y_centers = torch.linspace(-2.5, 2.5, nbins)
    z_centers = torch.linspace(-2.5, 2.5, nbins)
    xx, yy, zz = torch.meshgrid([x_centers, y_centers, z_centers])
    grid = torch.cat([xx.unsqueeze(-1), yy.unsqueeze(-1), zz.unsqueeze(-1)], -1)

    h3d: torch.Tensor = joint_density.pdf(grid) + damping
    h3d /= h3d.sum()
    return h3d


def hgr_cond(x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde]) -> np.ndarray:
    r"""An estimator of the function :math:`z\to HGR(x|z, y|z)`.

    Where HGR is the Hirschfeld-Gebelein-Renyi maximum correlation
    coefficient computed using Witsenhausen’s Characterization: HGR(x,y) is the second highest
    eigenvalue of the joint density on (x,y). We compute here the second eigenvalue on an empirical
    and discretized density estimated from the input data.

    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param z: A torch 1-D Tensor
    :param density: so far only kde is supported
    :returns: A torch 1-D Tensor of same size as Z.
        (0: independent, 1: linked by a deterministic equation)
    """
    damping = 1e-10
    h3d = _joint_3(x, y, z, density, damping=damping)
    marginal_xz = h3d.sum(dim=1).unsqueeze(1)
    marginal_yz = h3d.sum(dim=0).unsqueeze(0)
    Q = h3d / (torch.sqrt(marginal_xz) * torch.sqrt(marginal_yz))  # noqa: N806
    return np.array([torch.svd(Q[:, :, i])[1][1] for i in range(Q.shape[2])])


def chi_2_cond(
    x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde]
) -> torch.Tensor:
    r"""An estimator of the function :math:`z\to chi^2(x|z, y|z)`.

    Where :math:`\chi^2` is the :math:`\chi^2` divergence between the joint distribution on (x,y)
    and the product of marginals. This is know to be the square of an upper-bound on the
    Hirschfeld-Gebelein-Renyi maximum correlation coefficient. We compute it here on an empirical
    and discretized density estimated from the input data.

    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param z: A torch 1-D Tensor
    :param density: so far only kde is supported
    :returns: A torch 1-D Tensor of same size as Z. (0: independent)
    """
    damping = 0
    h3d = _joint_3(x, y, z, density, damping=damping)
    marginal_xz = h3d.sum(dim=1).unsqueeze(1)
    marginal_yz = h3d.sum(dim=0).unsqueeze(0)
    Q = h3d / (torch.sqrt(marginal_xz) * torch.sqrt(marginal_yz))  # noqa: N806
    return (Q**2).sum(dim=[0, 1]) - 1.0



"""Classes related to running algorithms in subprocesses."""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property
import json
from pathlib import Path
from tempfile import TemporaryDirectory, gettempdir
from typing import Any, Literal, Mapping, TypedDict, Union, final
from typing_extensions import Self, TypeAlias
import uuid

from ethicml.models.algorithm_base import SubprocessAlgorithmMixin
from ethicml.models.inprocess.in_algorithm import InAlgorithm
from ethicml.utility.data_structures import DataTuple, Prediction, TestTuple

__all__ = ["InAlgoArgs", "InAlgorithmSubprocess"]


class InAlgoRunArgs(TypedDict):
    """Base arguments for the ``run`` function of subprocess in-process methods."""

    mode: Literal["run"]
    predictions: str  # path to where the predictions should be stored
    # paths to the files with the data
    train: str
    test: str
    seed: int


class InAlgoFitArgs(TypedDict):
    """Base arguments for the ``fit`` function of subprocess in-process methods."""

    mode: Literal["fit"]
    train: str
    model: str  # path to where the model weights are stored
    seed: int


class InAlgoPredArgs(TypedDict):
    """Base arguments for the ``predict`` function of subprocess in-process methods."""

    mode: Literal["predict"]
    predictions: str
    test: str
    model: str


InAlgoArgs: TypeAlias = Union[InAlgoFitArgs, InAlgoPredArgs, InAlgoRunArgs]


@dataclass
class InAlgorithmSubprocess(SubprocessAlgorithmMixin, InAlgorithm, ABC):
    """In-Algorithm that uses a subprocess to run.

    :param dir: Directory to store the model.
    """

    dir: Path = field(default_factory=lambda: Path(gettempdir()))

    @cached_property  # needs to be cached because of the uuid4() call
    def model_path(self) -> Path:
        """Path to where the model with be stored."""
        name = self.name.replace(" ", "_")
        return self.dir.resolve(strict=True) / f"model_{name}_{uuid.uuid4()}.joblib"

    @final
    def fit(self, train: DataTuple, seed: int = 888) -> Self:
        """Fit Algorithm in a subprocess on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: Self, but trained.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = tmp_path / "train.npz"
            train.save_to_file(train_path)
            args: InAlgoFitArgs = {
                "mode": "fit",
                "train": str(train_path),
                "model": str(self.model_path),
                "seed": seed,
            }
            self.call_script(self._script_command(args))
            return self

    @final
    def predict(self, test: TestTuple) -> Prediction:
        """Make predictions in a subprocess on the given data.

        :param test: Data to evaluate on.
        :returns: Predictions on the test data.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            test_path = tmp_path / "test.npz"
            pred_path = tmp_path / "predictions.npz"
            test.save_to_file(test_path)
            args: InAlgoPredArgs = {
                "mode": "predict",
                "test": str(test_path),
                "predictions": str(pred_path),
                "model": str(self.model_path),
            }
            self.call_script(self._script_command(args))
            return Prediction.from_file(pred_path)

    @final
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        """Run Algorithm in a subprocess on the given data.

        :param train: Data tuple of the training data.
        :param test: Data to evaluate on.
        :param seed: Random seed for model initialization.
        :returns: Predictions on the test data.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = tmp_path / "train.npz"
            test_path = tmp_path / "test.npz"
            pred_path = tmp_path / "predictions.npz"
            train.save_to_file(train_path)
            test.save_to_file(test_path)
            args: InAlgoRunArgs = {
                "mode": "run",
                "train": str(train_path),
                "test": str(test_path),
                "predictions": str(pred_path),
                "seed": seed,
            }
            self.call_script(self._script_command(args))
            return Prediction.from_file(pred_path)

    @final
    def _script_command(self, in_algo_args: InAlgoArgs) -> list[str]:
        """Return the command that will run the script.

        The flag interface consists of two strings, both JSON strings: the general in-algo flags
        and then the more specific flags for the algorithm.

        :param in_algo_args: Arguments for the script.
        :returns: List of strings that will be passed to ``subprocess.run``.
        """
        interface = [
            json.dumps(in_algo_args, separators=(',', ':')),
            json.dumps(self._get_flags(), separators=(',', ':')),
        ]
        return self._get_path_to_script() + interface

    @abstractmethod
    def _get_path_to_script(self) -> list[str]:
        """Return arguments that are passed to the python executable."""

    @abstractmethod
    def _get_flags(self) -> Mapping[str, Any]:
        """Return flags that are used to configure this algorithm."""



"""For assessing ProbNeg."""
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import override

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric

__all__ = ["ProbNeg"]


@dataclass
class ProbNeg(CfmMetric):
    """Probability of negative prediction."""

    _name: ClassVar[str] = "prob_neg"

    @override
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, _, f_neg, _ = self._confusion_matrix(prediction=prediction, actual=actual)
        return (t_neg + f_neg) / prediction.hard.size



"""Data structure for all datasets that come with the framework."""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import auto
from pathlib import Path
import typing
from typing import ClassVar, Sequence, TypedDict, final
from typing_extensions import override

import pandas as pd
from ranzen.misc import StrEnum

from ethicml.common import ROOT_PATH
from ethicml.utility import DataTuple, undo_one_hot

from .util import (
    DiscFeatureGroups,
    LabelSpec,
    flatten_dict,
    from_dummies,
    label_spec_to_feature_list,
    single_col_spec,
)

__all__ = [
    "CSVDataset",
    "CSVDatasetDC",
    "Dataset",
    "FeatureOrder",
    "FeatureSplit",
    "LabelSpecsPair",
    "LegacyDataset",
    "StaticCSVDataset",
    "one_hot_encode_and_combine",
]


@dataclass
class LabelSpecsPair:
    """A pair of label specs.

    :param s: Spec for building the ``s`` label.
    :param y: Spec for building the ``y`` label.
    :param to_remove: List of feature groups that need to be removed because they are label building
        blocks. (Default: ``[]``)
    """

    s: LabelSpec
    y: LabelSpec
    to_remove: list[str] = field(default_factory=list)


class FeatureSplit(TypedDict):
    """A dictionary of the list of columns that belong to the feature groups."""

    x: list[str]
    s: list[str]
    y: list[str]


class FeatureOrder(StrEnum):
    """Order of features in the loaded datatuple."""

    cont_first = auto()
    """Continuous features first."""
    disc_first = auto()
    """Discrete features first."""


class Dataset(ABC):
    """Data structure that holds all the information needed to load a given dataset."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the dataset."""

    @abstractmethod
    def load(
        self, labels_as_features: bool = False, order: FeatureOrder = FeatureOrder.disc_first
    ) -> DataTuple:
        """Load dataset from its CSV file.

        :param labels_as_features: If ``True``, the s and y labels are included in the x features.
        :param order: Order of the columns in the dataframes. Can be ``disc_first`` or
            ``cont_first``. See :class:`FeatureOrder`.
        :returns: ``DataTuple`` with dataframes of features, labels and sensitive attributes.
        """

    @property
    @abstractmethod
    def continuous_features(self) -> list[str]:
        """Continuous features."""

    @property
    @abstractmethod
    def discrete_features(self) -> list[str]:
        """List of features that are discrete."""

    @abstractmethod
    def feature_split(self, order: FeatureOrder = FeatureOrder.disc_first) -> FeatureSplit:
        """Return an order features dictionary.

        This should have separate entries for the features, the labels and the
        sensitive attributes, but the x features are ordered so first are the discrete features,
        then the continuous.
        """

    @property
    @abstractmethod
    def disc_feature_groups(self) -> DiscFeatureGroups:
        """Return Dictionary of feature groups."""

    @abstractmethod
    def __len__(self) -> int:
        """Return number of elements in the dataset."""


class CSVDataset(Dataset, ABC):
    """Dataset class that loads data from CSV files."""

    discard_non_one_hot: ClassVar[bool] = False
    """If some entries in s or y are not correctly one-hot encoded, discard those."""
    map_to_binary: ClassVar[bool] = False
    """If True, convert labels from {-1, 1} to {0, 1}."""

    @abstractmethod
    def get_label_specs(self) -> LabelSpecsPair:
        """Label specs and discrete feature groups that have to be removed from ``x``."""

    @abstractmethod
    def get_num_samples(self) -> int:
        """Number of samples in the dataset."""

    @abstractmethod
    def get_filename_or_path(self) -> str | Path:
        """Filename of CSV files containing the data."""

    @property
    @abstractmethod
    def load_discrete_only(self) -> bool:
        """Whether to only load discrete features."""

    @property
    @abstractmethod
    def invert_sens_attr(self) -> bool:
        """Whether to invert the sensitive attribute."""

    @property
    @abstractmethod
    def unfiltered_disc_feat_groups(self) -> DiscFeatureGroups:
        """Discrete feature groups, including features for the labels."""

    @property
    @final
    def filepath(self) -> Path:
        """Filepath from which to load the data."""
        file_name_or_path = self.get_filename_or_path()
        if isinstance(file_name_or_path, Path):
            return file_name_or_path
        else:
            return ROOT_PATH / "data" / "csvs" / file_name_or_path

    @property
    @final
    def sens_attrs(self) -> list[str]:
        """Get the list of sensitive attributes."""
        return label_spec_to_feature_list(self.get_label_specs().s)

    @property
    @final
    def class_labels(self) -> list[str]:
        """Get the list of class labels."""
        return label_spec_to_feature_list(self.get_label_specs().y)

    @override
    def feature_split(self, order: FeatureOrder = FeatureOrder.disc_first) -> FeatureSplit:
        if self.load_discrete_only:
            x = self.discrete_features
        elif order is FeatureOrder.disc_first:
            x = self.discrete_features + self.continuous_features
        else:
            x = self.continuous_features + self.discrete_features
        label_specs = self.get_label_specs()
        return {
            "x": x,
            "s": label_spec_to_feature_list(label_specs.s),
            "y": label_spec_to_feature_list(label_specs.y),
        }

    @property
    def discrete_features(self) -> list[str]:
        """List of features that are discrete."""
        return flatten_dict(self.disc_feature_groups)

    @property
    def disc_feature_groups(self) -> DiscFeatureGroups:
        """Return Dictionary of feature groups, without s and y labels."""
        dfgs = self.unfiltered_disc_feat_groups
        # select those feature groups that are not for the x and y labels
        to_remove = self.get_label_specs().to_remove
        assert all(group in dfgs for group in to_remove), "can't remove all groups"
        return {group: v for group, v in dfgs.items() if group not in to_remove}

    @override
    def __len__(self) -> int:
        return self.get_num_samples()

    @override
    def load(
        self, labels_as_features: bool = False, order: FeatureOrder = FeatureOrder.disc_first
    ) -> DataTuple:
        dataframe: pd.DataFrame = pd.read_csv(self.filepath, nrows=self.get_num_samples())
        assert isinstance(dataframe, pd.DataFrame)

        feature_split = self.feature_split(order=order)
        if labels_as_features:
            feature_split_x = feature_split["x"] + feature_split["s"] + feature_split["y"]
        else:
            feature_split_x = feature_split["x"]

        for feature_group in self.unfiltered_disc_feat_groups.values():
            dataframe = _generate_complementary_column(dataframe, feature_group)

        x_data = dataframe[feature_split_x]
        s_df = dataframe[feature_split["s"]]
        y_df = dataframe[feature_split["y"]]

        if self.map_to_binary:
            s_df = (s_df + 1) // 2  # map from {-1, 1} to {0, 1}
            y_df = (y_df + 1) // 2  # map from {-1, 1} to {0, 1}

        if self.invert_sens_attr:
            assert s_df.nunique().to_numpy()[0] == 2, "s must be binary"
            s_df = 1 - s_df

        label_specs = self.get_label_specs()

        # the following operations remove rows if a label group is not properly one-hot encoded
        s_data, s_mask = one_hot_encode_and_combine(s_df, label_specs.s, self.discard_non_one_hot)
        if s_mask is not None:
            x_data = x_data.loc[s_mask].reset_index(drop=True)
            s_data = s_data.loc[s_mask].reset_index(drop=True)
            y_df = y_df.loc[s_mask].reset_index(drop=True)
        y_data, y_mask = one_hot_encode_and_combine(y_df, label_specs.y, self.discard_non_one_hot)
        if y_mask is not None:
            x_data = x_data.loc[y_mask].reset_index(drop=True)
            s_data = s_data.loc[y_mask].reset_index(drop=True)
            y_data = y_data.loc[y_mask].reset_index(drop=True)

        return DataTuple.from_df(x=x_data, s=s_data, y=y_data, name=self.name)

    @typing.no_type_check
    def load_aif(self):  # Returns aif.360 Standard Dataset
        """Load the dataset as an AIF360 dataset.

        Experimental.
        Requires the aif360 library.

        Ignores the type check as the return type is not yet defined.

        """
        from aif360.datasets import StandardDataset

        data = self.load()
        disc_feature_groups = self.disc_feature_groups
        data_collapsed = from_dummies(data.x, disc_feature_groups)
        data = data.replace(x=data_collapsed)
        df = data.data

        return StandardDataset(
            df=df,
            label_name=data.y_column,
            favorable_classes=lambda x: x > 0,
            protected_attribute_names=[data.s_column],
            privileged_classes=[lambda x: x == 1],
            categorical_features=disc_feature_groups.keys(),
        )


def _generate_complementary_column(
    dataframe: pd.DataFrame, feature_group: list[str]
) -> pd.DataFrame:
    """Check whether we have to generate some complementary columns for binary features.

    This happens when we have, for example, several races: race-asian-pac-islander, etc, but we want
    to have an attribute called "race_other" that summarizes them all. Now, the problem is that this
    cannot be done before this point, because only here have we actually loaded the data. So, we
    have to do it here, with all the information we can piece together.
    """
    for feature in feature_group:
        if feature in dataframe.columns:
            continue  # nothing to do
        missing_feature = feature

        existing_features = [other for other in feature_group if other in dataframe.columns]
        assert existing_features, f"no column from this feature group seems to exist: {feature}"
        assert len(existing_features) == len(feature_group) - 1, "at most 1 feature missing"
        # the dummy feature is the inverse of the existing feature
        or_combination = dataframe[existing_features[0]] == 1
        for other in existing_features[1:]:
            or_combination |= dataframe[other] == 1
        inverse: pd.Series = 1 - or_combination
        dataframe = pd.concat([dataframe, inverse.to_frame(name=missing_feature)], axis="columns")
    return dataframe


def one_hot_encode_and_combine(
    attributes: pd.DataFrame, label_spec: LabelSpec, discard_non_one_hot: bool
) -> tuple[pd.Series, pd.Series | None]:
    """Construct a new label according to the given :class:`~ethicml.data.util.LabelSpec`.

    This function is at the heart of the label spec API in EthicML.

    :param attributes: DataFrame containing the attributes.
    :param label_spec: A label spec.
    :param discard_non_one_hot: If ``True``, a mask is returned which masks out all rows which are
        not properly one-hot (i.e., either all classes are 0 or more than one is 1).
    :returns: A tuple of a Series with the new labels and -- if ``discard_non_one_hot`` is ``True``
        -- a mask for filtering out the rows that were not properly one-hot.
    """
    mask = None  # the mask is needed when we have to discard samples

    # create a Series of zeroes with the same length as the dataframe
    combination: pd.Series = pd.Series(0, index=range(len(attributes)), name=",".join(label_spec))

    for name, spec in label_spec.items():
        if len(spec.columns) > 1:  # data is one-hot encoded
            raw_values = attributes[spec.columns]
            if discard_non_one_hot:
                # only use those samples where exactly one of the specified attributes is true
                mask = raw_values.sum(axis="columns") == 1
            else:
                assert (raw_values.sum(axis="columns") == 1).all(), f"{name} is not one-hot"
            values = undo_one_hot(raw_values)
        else:
            values = attributes[spec.columns[0]]
        combination += spec.multiplier * values
    return combination, mask


@dataclass
class CSVDatasetDC(CSVDataset, ABC):
    """Dataset that uses the default load function."""

    discrete_only: bool = False
    invert_s: bool = False

    @property
    def load_discrete_only(self) -> bool:
        """Whether to only load discrete features."""
        return self.discrete_only

    @property
    def invert_sens_attr(self) -> bool:
        """Whether to invert the sensitive attribute."""
        return self.invert_s


@dataclass
class StaticCSVDataset(CSVDatasetDC, ABC):
    """Dataset whose size and file location does not depend on constructor arguments.

    :example:
        How to subclass this:

        .. code:: python

            @dataclass
            class Toy(StaticCSVDataset):
                '''Dataset with toy data for testing.'''

                num_samples: ClassVar[int] = 400
                csv_file: ClassVar[str] = "toy.csv"

                @property
                def name(self) -> str:
                    return "Toy"

                def get_label_specs(self) -> LabelSpecsPair:
                    return LabelSpecsPair(
                        s=single_col_spec("sens"), y=single_col_spec("class")
                    )

                @property
                def unfiltered_disc_feat_groups(self) -> DiscFeatureGroups:
                    return {"disc_1": ["a_1", "a_2", "a_3"], "disc_2": ["b_1", "b_2"]}

                @property
                def continuous_features(self) -> list[str]:
                    return ["c1", "c2"]
    """

    num_samples: ClassVar[int] = 0
    csv_file: ClassVar[str] = "<overwrite me>"

    @override
    def get_num_samples(self) -> int:
        return self.num_samples

    @override
    def get_filename_or_path(self) -> str | Path:
        return self.csv_file


@dataclass(init=False)
class LegacyDataset(CSVDataset):
    """Dataset base class.

    This base class is considered legacy now. Please use :class:`CSVDatasetDC` or
    :class:`StaticCSVDataset` instead.
    """

    discrete_only: bool = False
    invert_s: bool = False

    def __init__(
        self,
        *,
        name: str,
        filename_or_path: str | Path,
        features: Sequence[str],
        cont_features: Sequence[str],
        sens_attr_spec: str | LabelSpec,
        class_label_spec: str | LabelSpec,
        num_samples: int,
        s_feature_groups: Sequence[str] | None = None,
        class_feature_groups: Sequence[str] | None = None,
        discrete_feature_groups: dict[str, list[str]] | None = None,
    ) -> None:
        self._name = name
        self._features = features
        self._sens_attr_spec = sens_attr_spec
        self._class_label_spec = class_label_spec
        self._num_samples = num_samples
        self._s_prefix = [] if s_feature_groups is None else s_feature_groups
        self._class_label_prefix = [] if class_feature_groups is None else class_feature_groups
        if discrete_feature_groups is not None:
            self._unfiltered_disc_feat_groups = discrete_feature_groups
        else:
            self._unfiltered_disc_feat_groups = {
                feat: [feat] for feat in self._features if feat not in cont_features
            }
        self._raw_file_name_or_path = filename_or_path
        self._cont_features = list(cont_features)

    @property
    @override
    def unfiltered_disc_feat_groups(self) -> DiscFeatureGroups:
        return self._unfiltered_disc_feat_groups

    @property
    @override
    def continuous_features(self) -> list[str]:
        return self._cont_features

    @property
    @override
    def name(self) -> str:
        return self._name

    @override
    def get_label_specs(self) -> LabelSpecsPair:
        s_spec = self._sens_attr_spec
        y_spec = self._class_label_spec
        return LabelSpecsPair(
            s=single_col_spec(s_spec) if isinstance(s_spec, str) else s_spec,
            y=single_col_spec(y_spec) if isinstance(y_spec, str) else y_spec,
            to_remove=list(self._s_prefix) + list(self._class_label_prefix),
        )

    @override
    def get_num_samples(self) -> int:
        return self._num_samples

    @override
    def get_filename_or_path(self) -> str | Path:
        return self._raw_file_name_or_path

    @property
    def load_discrete_only(self) -> bool:
        """Whether to only load discrete features."""
        return self.discrete_only

    @property
    def invert_sens_attr(self) -> bool:
        """Whether to invert the sensitive attribute."""
        return self.invert_s



"""HGR Classifier.

Paper: "Fairness-Aware Learning for Continuous Attributes and Treatments",
J. Mary, C. Calauzènes, N. El Karoui, ICML 2019.
Based on: https://github.com/criteo-research/continuous-fairness
http://proceedings.mlr.press/v97/mary19a/mary19a.pdf
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import TypedDict
from typing_extensions import override

from ethicml.models.inprocess.in_subprocess import InAlgorithmSubprocess
from ethicml.utility import HyperParamType, ModelType

__all__ = ["HGR", "HgrArgs"]


class HgrArgs(TypedDict):
    """Args for the HGR implementation."""

    lr: float
    epochs: int
    mu: float
    batch_size: int
    model_type: str


@dataclass
class HGR(InAlgorithmSubprocess):
    """HGR Method."""

    lr: float = 1e-3
    epochs: int = 50
    mu: float = 0.98
    batch_size: int = 128
    model_type: ModelType = ModelType.deep

    @override
    def _get_path_to_script(self) -> list[str]:
        return ["-m", "ethicml.implementations.hgr_method"]

    @override
    def _get_flags(self) -> HgrArgs:
        return {
            "lr": self.lr,
            "epochs": self.epochs,
            "mu": self.mu,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
        }

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {
            "lr": self.lr,
            "epochs": self.epochs,
            "mu": self.mu,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
        }

    @property
    @override
    def name(self) -> str:
        return f"HGR {self.model_type}_model"



"""Definitions of algorithms."""
from .algorithm_base import *
from .inprocess.adv_debiasing import *
from .inprocess.agarwal_reductions import *
from .inprocess.blind import *
from .inprocess.fair_dummies import *
from .inprocess.fairness_wo_demographics import *
from .inprocess.hgr import *
from .inprocess.in_algorithm import *
from .inprocess.in_subprocess import *
from .inprocess.installed_model import *
from .inprocess.kamiran import *
from .inprocess.kamishima import *
from .inprocess.logistic_regression import *
from .inprocess.majority import *
from .inprocess.manual import *
from .inprocess.mlp import *
from .inprocess.oracle import *
from .inprocess.svm import *
from .inprocess.zafar import *
from .postprocess.dp_flip import *
from .postprocess.hardt import *
from .postprocess.post_algorithm import *
from .preprocess.beutel import *
from .preprocess.calders import *
from .preprocess.pre_algorithm import *
from .preprocess.pre_subprocess import *
from .preprocess.upsampler import *
from .preprocess.vfae import *
from .preprocess.zemel import *



"""Installable model.

This is a kind of complicated model, but it's incredibly useful.
Say you find a paper from a few years ago with code. It's not unreasonable that there might
be dependency clashes, python clashes, clashes galore. This approach downloads a model, runs it
in its own venv and makes everyone happy.
"""
from __future__ import annotations
from abc import ABC
import os
from pathlib import Path
import shutil
import subprocess
import sys
from typing_extensions import override

import git.cmd

from ethicml.models.inprocess.in_algorithm import InAlgorithm

from ..algorithm_base import SubprocessAlgorithmMixin

__all__ = ["InstalledModel"]


class InstalledModel(SubprocessAlgorithmMixin, InAlgorithm, ABC):
    """The model that does the magic.

    Download code from given URL and create Pip environment with Pipfile found in the code.

    :param name: Name of the model.
    :param dir_name: Where to download the code to (can be chosen freely).
    :param top_dir: Top directory of the repository where the Pipfile can be found (this is usually
        simply the last part of the repository URL).
    :param url: URL of the repository. (Default: None)
    :param executable: Path to a Python executable. (Default: None.
    :param use_pdm: If ``True``, will try to use pdm instead of pipenv. (Default: False)
    """

    def __init__(
        self,
        name: str,
        dir_name: str,
        top_dir: str,
        url: str | None = None,
        executable: list[str] | None = None,
        use_pdm: bool = False,
    ):
        # QUESTION: do we really need `store_dir`? we could also just clone the code into "."
        self._store_dir: Path = Path(".") / dir_name  # directory where code and venv are stored
        self._top_dir: str = top_dir

        if url is not None:
            # download code
            self._clone_directory(url)

        if executable is None:
            # create virtual environment
            self.__executable = self._create_venv(use_pdm=use_pdm)
        else:
            self.__executable = executable
        self.__name = name

    @property
    @override
    def name(self) -> str:
        return self.__name

    @property
    def _code_path(self) -> Path:
        """Path to where the code of the model is located."""
        return self._store_dir / self._top_dir

    @property
    def executable(self) -> list[str]:
        """Python executable from the virtualenv associated with the model."""
        return self.__executable

    def _clone_directory(self, url: str) -> None:
        """Clones the repo from `url` into `self._store_dir`.

        :param url: URL of the repository.
        """
        if not self._store_dir.exists():
            self._store_dir.mkdir()
            git.cmd.Git(self._store_dir).clone(url)

    def _create_venv(self, use_pdm: bool) -> list[str]:
        """Create a venv based on the Pipfile in the repository.

        :param use_pdm: Whether to use pdm instead of pipenv.
        :returns: Shell command to execute python scripts.
        :raises RuntimeError: If the required package manager isn't found.
        """
        if use_pdm:  # use pdm instead of pipenv
            pypackages_dir = self._code_path / "__pypackages__"
            if not pypackages_dir.exists():
                if shutil.which("pdm") is None:
                    raise RuntimeError("pdm must be installed")
                environ = os.environ.copy()
                environ["PDM_USE_VENV"] = "0"
                subprocess.run(
                    ["pdm", "install", "--no-self"], env=environ, check=True, cwd=self._code_path
                )
            return ["pdm", "run", "python"]

        venv_directory = self._code_path.resolve() / ".venv"
        if not venv_directory.exists():
            environ = os.environ.copy()
            environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
            environ["PIPENV_VENV_IN_PROJECT"] = "true"
            environ["PIPENV_YES"] = "true"
            environ["PIPENV_PIPFILE"] = str(self._code_path / "Pipfile")

            subprocess.run([sys.executable, "-m", "pipenv", "install"], env=environ, check=True)
        return [str(venv_directory / "bin" / "python")]

    def remove(self) -> None:
        """Remove the directory that we created in :meth:`_clone_directory()`."""
        try:
            shutil.rmtree(self._store_dir)
        except OSError as excep:
            print(f"Error: {excep.filename} - {excep.strerror}.")



"""Wrapper for calling Kamishima model."""
from __future__ import annotations
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import NamedTuple
from typing_extensions import override

import numpy as np
import pandas as pd

from ethicml.models.inprocess.installed_model import InstalledModel
from ethicml.utility import DataTuple, HyperParamType, Prediction, TestTuple

__all__ = ["Kamishima"]


class _FitInfo(NamedTuple):
    """Information that is stored after the model has been fit."""

    min_class_label: int
    model_path: Path


class Kamishima(InstalledModel):
    """Model that calls Kamishima's code.

    Based on Algo-Fairness
    https://github.com/algofairness/fairness-comparison/blob/master/fairness/algorithms/kamishima/KamishimaAlgorithm.py

    :param eta: Tolerance.
    """

    def __init__(self, *, eta: float = 1.0):
        super().__init__(
            name="Kamishima",
            dir_name="kamishima",
            url="https://github.com/wearepal/kamfadm.git",
            top_dir="kamfadm",
            use_pdm=True,
        )
        self.eta = eta
        self._fit_info: _FitInfo | None = None

    @property
    @override
    def hyperparameters(self) -> HyperParamType:
        return {"eta": self.eta}

    @override
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            fit_info = self._fit(train, tmp_path)
            return self._predict(test, fit_info, tmp_path)

    @override
    def fit(self, train: DataTuple, seed: int = 888) -> Kamishima:
        with TemporaryDirectory() as tmpdir:
            self._fit_info = self._fit(train, Path(tmpdir), model_dir=self._code_path)
        return self

    def _fit(self, train: DataTuple, tmp_path: Path, model_dir: Path | None = None) -> _FitInfo:
        train_path = tmp_path / "train.txt"
        _create_file_in_kamishima_format(train, train_path)
        min_class_label: int = train.y.min()
        model_path = (model_dir if model_dir is not None else tmp_path) / "model"

        script = self._code_path / "train_pr.py"
        cmd = [script, "-e", self.eta, "-i", train_path, "-o", model_path, "--quiet"]
        self.call_script([str(e) for e in cmd])
        return _FitInfo(min_class_label=min_class_label, model_path=model_path)

    @override
    def predict(self, test: TestTuple) -> Prediction:
        assert self._fit_info is not None, "call fit() before calling predict()"
        with TemporaryDirectory() as tmpdir:
            return self._predict(test, self._fit_info, Path(tmpdir))

    def _predict(self, test: TestTuple, fit_info: _FitInfo, tmp_path: Path) -> Prediction:
        test_path = tmp_path / "test.txt"
        _create_file_in_kamishima_format(test, test_path)
        output_path = str(tmp_path / "output.txt")
        script = self._code_path / "predict_lr.py"
        cmd = [script, "-i", test_path, "-m", fit_info.model_path, "-o", output_path, "--quiet"]
        self.call_script([str(e) for e in cmd])
        output = np.loadtxt(output_path)
        predictions = output[:, 1].astype(np.float32)
        # except RuntimeError:
        #     predictions = np.ones_like(test.y.to_numpy())

        to_return = pd.Series(predictions)
        to_return = to_return.astype(int)

        if (min_val := to_return.min()) != to_return.max():
            to_return = to_return.replace(min_val, fit_info.min_class_label)
        return Prediction(hard=to_return)


def _create_file_in_kamishima_format(data: DataTuple | TestTuple, file_path: Path) -> None:
    """Create a text file with the data.

    :param data: Data to write to the file.
    :param file_path: Path to the file.
    """
    if isinstance(data, DataTuple):
        result = pd.concat([data.x, data.s, data.y], axis="columns").to_numpy().astype(np.float64)
    else:
        zeros = pd.DataFrame([0 for _ in range(data.x.shape[0])], columns=["y"])
        result = pd.concat([data.x, data.s, zeros], axis="columns").to_numpy().astype(np.float64)
    np.savetxt(file_path, result)



"""Method for calculating the HSIC - an independence criterion.

a score of 0 denotes independence
"""
from __future__ import annotations
from dataclasses import dataclass
import math
from typing import ClassVar
from typing_extensions import override

import numpy as np
from numpy.random import RandomState

from ethicml.metrics.metric import MetricStaticName
from ethicml.utility import EvalTuple, Prediction

__all__ = ["Hsic"]


def hsic(
    prediction: np.ndarray, label: np.ndarray, sigma_first: float, sigma_second: float
) -> float:
    """Calculate the HSIC value."""
    xx_gram = np.array(np.matmul(np.expand_dims(prediction, 1), np.expand_dims(prediction, 1).T))
    yy_gram = np.array(np.matmul(np.expand_dims(label, 1), np.expand_dims(label, 1).T))

    x_sqnorms = np.diag(xx_gram)
    y_sqnorms = np.diag(yy_gram)

    def exp_r(x: np.ndarray) -> np.ndarray:
        return np.expand_dims(x, 0)

    def exp_c(x: np.ndarray) -> np.ndarray:
        return np.expand_dims(x, 1)

    gamma_first = 1.0 / (2 * sigma_first**2)
    gamma_second = 1.0 / (2 * sigma_second**2)
    # use the second binomial formula
    kernel_xx = np.exp(-gamma_first * (-2 * xx_gram + exp_c(x_sqnorms) + exp_r(x_sqnorms)))
    kernel_yy = np.exp(-gamma_second * (-2 * yy_gram + exp_c(y_sqnorms) + exp_r(y_sqnorms)))

    kernel_xx_mean = np.mean(kernel_xx)
    kernel_yy_mean = np.mean(kernel_yy)

    h_k = kernel_xx - kernel_xx_mean
    h_l = kernel_yy - kernel_yy_mean

    num = float(kernel_yy.shape[0])
    h_kf = h_k / (num - 1)
    h_lf = h_l / (num - 1)

    # biased estimate
    hsic_value = float(np.trace(np.matmul(h_kf.T, h_lf)))
    return hsic_value


@dataclass
class Hsic(MetricStaticName):
    """We add the ability to take the average of hsic score.

    As for larger datasets it will kill your machine.
    """

    seed: int = 888
    _name: ClassVar[str] = "HSIC"
    apply_per_sensitive: ClassVar[bool] = False

    @override
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        preds = prediction.hard.to_numpy()[:, np.newaxis]
        sens_labels = actual.s.to_numpy()[:, np.newaxis]

        batchs_size = 5000

        together = np.hstack((preds, sens_labels)).transpose()

        random = RandomState(seed=self.seed)
        col_idx = random.permutation(together.shape[1])

        together = np.take(together, col_idx, axis=1)

        prediction_shuffled = together[0]
        label_shuffled = together[1]

        num_batches_float = preds.shape[0] / batchs_size
        num_batches: int = int(math.ceil(num_batches_float))

        batches = []

        start = 0

        for _ in range(num_batches):
            end = start + batchs_size

            preds_to_test = prediction_shuffled[start:end]
            labels_to_test = label_shuffled[start:end]

            batches.append(hsic(preds_to_test, labels_to_test, 0.7, 0.5))

            start += batchs_size

        return np.array(batches).mean().item()



"""Create plots of a dataset."""
from __future__ import annotations
import itertools
from pathlib import Path
from typing import Any, Callable, List, Literal, cast

from matplotlib import figure, legend
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from ethicml.metrics.metric import Metric
from ethicml.utility import DataTuple, Results

from .common import DataEntry, LegendType, PlotDef, PlotType, errorbox, scatter

__all__ = [
    "plot_results",
    "single_plot",
    "save_2d_plot",
    "save_label_plot",
    "save_jointplot",
    "save_multijointplot",
]

MARKERS = ["s", "p", "P", "*", "+", "x", "o", "v"]


def _maybe_tsne(data: DataTuple) -> tuple[pd.DataFrame, str, str]:
    columns = data.x.columns

    if len(columns) > 2:
        from sklearn.manifold import TSNE

        tsne_embeddings = TSNE(n_components=2, random_state=0).fit_transform(data.x)
        amalgamated = pd.concat(
            [pd.DataFrame(tsne_embeddings, columns=["tsne1", "tsne2"]), data.s, data.y],
            axis="columns",
        )
        x1_name = "tsne1"
        x2_name = "tsne2"
    else:
        amalgamated = data.data
        x1_name = f"{columns[0]}"
        x2_name = f"{columns[1]}"
    return amalgamated, x1_name, x2_name


def save_2d_plot(data: DataTuple, filepath: str) -> None:
    """Make 2D plot."""
    file_path = Path(filepath)

    amalgamated, x1_name, x2_name = _maybe_tsne(data)

    plot = sns.scatterplot(  # type: ignore[attr-defined]
        x=x1_name,
        y=x2_name,
        hue=data.y.name,
        palette="Set2",
        data=amalgamated,
        style=data.s.name,
        legend="full",
    )

    file_path.parent.mkdir(exist_ok=True)
    plot.figure.savefig(file_path)
    plt.clf()


def save_jointplot(data: DataTuple, filepath: str, dims: tuple[int, int] = (0, 1)) -> None:
    """Make joint plot."""
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.y], axis="columns")

    plot = sns.jointplot(  # type: ignore[attr-defined]
        x=columns[dims[0]], y=columns[dims[1]], data=amalgamated, kind="kde"
    )

    file_path.parent.mkdir(exist_ok=True)
    plot.savefig(file_path)
    plt.clf()


def _multivariate_grid(
    col_x: str,
    col_y: str,
    sens_col: str,
    outcome_col: str,
    df: pd.DataFrame,
    scatter_alpha: float = 0.5,
) -> None:
    def colored_scatter(x: Any, y: Any, c: str | None = None) -> Callable[[Any], None]:
        def scatter(*args: Any, **kwargs: Any) -> None:
            args = (x, y)
            if c is not None:
                kwargs['c'] = c
            kwargs['alpha'] = scatter_alpha
            plt.scatter(*args, **kwargs)

        return scatter

    sns.set_palette("husl")  # type: ignore[attr-defined]

    g = sns.JointGrid(x=col_x, y=col_y, data=df)  # type: ignore[attr-defined]
    color = None
    legends = []
    for name, df_group in df.groupby([sens_col, outcome_col]):
        legends.append(f"S={name[0]}, Y={name[1]}")
        g.plot_joint(colored_scatter(df_group[col_x], df_group[col_y], color))
        sns.distplot(  # type: ignore[attr-defined]
            df_group[col_x].to_numpy(), ax=g.ax_marg_x, color=color
        )
        sns.distplot(  # type: ignore[attr-defined]
            df_group[col_y].to_numpy(), ax=g.ax_marg_y, vertical=True
        )
    # Do also global Hist:
    # sns.distplot(df[col_x].values, ax=g.ax_marg_x, color='grey')
    # sns.distplot(df[col_y].values.ravel(), ax=g.ax_marg_y, color='grey', vertical=True)
    plt.legend(legends)


def save_multijointplot(data: DataTuple, filepath: str) -> None:
    """Make joint plot."""
    file_path = Path(filepath)
    data.x.columns

    amalgamated, x1_name, x2_name = _maybe_tsne(data)

    _multivariate_grid(
        col_x=x1_name,
        col_y=x2_name,
        sens_col=str(data.s.name),
        outcome_col=str(data.y.name),
        df=amalgamated,
    )

    file_path.parent.mkdir(exist_ok=True)
    plt.savefig(file_path)
    plt.clf()


def make_gif(files: list[str], name: str = "movie") -> None:
    """Make GIF."""
    import imageio

    images = [imageio.imread(filename) for filename in files]
    imageio.mimsave(f"{name}.gif", images)


def save_label_plot(data: DataTuple, filename: str) -> None:
    """Make label plot."""
    file_path = Path(filename)

    # Only consider 1 sens attr for now
    s_values = data.s.value_counts() / data.s.count()

    s_0_val = s_values[0]
    s_1_val = s_values[1]

    s_0_label = s_values.index.min()
    s_1_label = s_values.index.max()

    y_s0 = data.y[data.s == 0].value_counts() / data.y[data.s == 0].count()
    y_s1 = data.y[data.s == 1].value_counts() / data.y[data.s == 1].count()

    y_0_label = y_s0.index[0]
    y_1_label = y_s0.index[1]

    plt.style.use("seaborn-pastel")
    # plt.xkcd()

    fig, plot = plt.subplots()

    quadrant1 = plot.bar(
        0,
        height=y_s0[y_0_label] * 100,  # type: ignore[index]
        width=s_0_val * 100,
        align="edge",
        edgecolor="black",
        color="C0",
    )
    quadrant2 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_0_label] * 100,  # type: ignore[index]
        width=s_1_val * 100,
        align="edge",
        edgecolor="black",
        color="C1",
    )
    quadrant3 = plot.bar(
        0,
        height=y_s0[y_1_label] * 100,  # type: ignore[index]
        width=s_0_val * 100,
        bottom=y_s0[y_0_label] * 100,  # type: ignore[index]
        align="edge",
        edgecolor="black",
        color="C2",
    )
    quadrant4 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_1_label] * 100,  # type: ignore[index]
        width=s_1_val * 100,
        bottom=y_s1[y_0_label] * 100,  # type: ignore[index]
        align="edge",
        edgecolor="black",
        color="C3",
    )

    plot.set_ylim(0, 100)
    plot.set_xlim(0, 100)
    plot.set_ylabel(f"Percent {data.y.name}=y")
    plot.set_xlabel(f"Percent {data.s.name}=s")
    plot.set_title("Dataset Composition by class and sensitive attribute")

    plot.legend(
        [quadrant1, quadrant2, quadrant3, quadrant4],
        [
            f"y={y_0_label}, s={s_0_label}",
            f"y={y_0_label}, s={s_1_label}",
            f"y={y_1_label}, s={s_0_label}",
            f"y={y_1_label}, s={s_1_label}",
        ],
    )

    file_path.parent.mkdir(exist_ok=True)
    fig.savefig(file_path)  # type: ignore
    fig.clf()


def single_plot(
    plot: plt.Axes,
    results: Results,
    xaxis: tuple[str, str],
    yaxis: tuple[str, str],
    dataset: str,
    transform: str | None,
    *,
    ptype: PlotType = "box",
    legend_pos: LegendType | None = "outside",
    legend_yanchor: float = 1.0,
    markersize: int = 6,
    alternating_style: bool = True,
    include_nan_entries: bool = False,
) -> None | Literal[False] | legend.Legend:
    """Provide the functionality of the individual plotting functions through a nice interface.

    This function can also be used to create figures with multiple plots on them, because it does
    not generate a Figure object itself.

    :param plot: Plot object
    :param results: DataFrame with the data
    :param xaxis: Name of column that's plotted on the x-axis.
    :param yaxis: Name of column that's plotted on the y-axis.
    :param dataset: string that identifies the dataset
    :param transform: string that identifies the preprocessing method, or None
    :param ptype: plot type (Default: "box")
    :param legend_pos: position of the legend (or None for no legend) (Default: "outside")
    :param legend_yanchor: position in the vertical direction where the legend should begin (Default
        value = 1.0)
    :param markersize: size of marker (Default: 6)
    :param alternating_style: If True, entries for scatter plots are done in alternating style
        (Default: True)
    :param include_nan_entries: If True, entries with NaNs still appear in the legend (Default value
        = False)
    :returns: The legend object if something was plotted; False otherwise.
    :raises ValueError: If ``ptype`` has an unsupported value.
    """
    mask_for_dataset = results.index.get_level_values("dataset") == dataset
    if transform is not None:
        transforms: list[str] = [transform]
    else:
        transforms = [str(t) for t in results.index.to_frame()["transform"].unique()]

    entries: list[DataEntry] = []
    count = 0
    for model in results.index.to_frame()["model"].unique():
        mask_for_model = results.index.get_level_values("model") == model
        for transform_ in transforms:
            mask_for_transform = results.index.get_level_values("transform") == transform_
            data = results.loc[mask_for_dataset & mask_for_model & mask_for_transform]
            if (
                data[[xaxis[0], yaxis[0]]].empty or data[[xaxis[0], yaxis[0]]].isna().any().any()
            ) and not include_nan_entries:
                continue  # this entry has missing values
            model_label = f"{model} ({transform_})" if transform_ != "no_transform" else str(model)
            entries.append(DataEntry(model_label, data, (not alternating_style) or count % 2 == 0))
            count += 1

    if not entries:
        return False  # nothing to plot

    title = f"{dataset}, {transform}" if transform is not None else dataset
    plot_def = PlotDef(
        title=title, entries=entries, legend_pos=legend_pos, legend_yanchor=legend_yanchor
    )
    if ptype == "box":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=False)
    if ptype == "cross":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=True)
    if ptype == "scatter":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=False)
    if ptype == "line":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=True)
    raise ValueError(f"Unknown plot type '{ptype}'")


def plot_results(
    results: Results,
    metric_y: str | Metric,
    metric_x: str | Metric,
    ptype: PlotType = "box",
    save: bool = True,
    dpi: int = 300,
    transforms_separately: bool = True,
) -> list[tuple[figure.Figure, plt.Axes]]:
    """Plot the given result with boxes that represent mean and standard deviation.

    :param results: A DataFrame that already contains the values of the metrics.
    :param metric_y: A Metric object or a column name that defines which metric to plot on the
        y-axis.
    :param metric_x: A Metric object or a column name that defines which metric to plot on the
        x-axis.
    :param ptype: Plot type (Default: "box")
    :param save: If ``True``, save the plot as a PDF. (Default: True)
    :param dpi: DPI of the plots (Default: 300)
    :param transforms_separately: If ``True``, each transform gets its own plot (Default: True)
    :returns: A list of all figures and plots.
    :raises ValueError: If no columns matching ``metric_y`` and ``metric_x`` are found.
    """
    directory = Path(".") / "plots"
    directory.mkdir(exist_ok=True)

    def _get_columns(metric: Metric) -> list[str]:
        cols = [col for col in cast(List[str], results.columns) if metric.name in col]
        if not cols:
            raise ValueError(f'No matching columns found for Metric "{metric.name}".')
        # if there are multiple matches, then the metric was `per_sensitive_attribute`. In this
        # case, we *only* want ratios and differences; not the plain result
        if len(cols) > 1:
            cols = [col for col in cols if ("-" in col) or ("÷" in col)]
        return cols

    if isinstance(metric_x, str):
        if metric_x not in results.columns:
            raise ValueError(f'No column named "{metric_x}".')
        cols_x = [metric_x]
    else:
        # if the metric is given as a Metric object, look for matching columns
        cols_x = _get_columns(metric_x)

    if isinstance(metric_y, str):
        if metric_y not in results.columns:
            raise ValueError(f'No column named "{metric_y}".')
        cols_y = [metric_y]
    else:
        cols_y = _get_columns(metric_y)

    # generate the Cartesian product of `cols_x` and `cols_y`; i.e. all possible combinations
    # this preserves the order of x and y
    possible_pairs = list(itertools.product(cols_x, cols_y))

    transforms: list[str | None]
    if transforms_separately:
        transforms = [str(t) for t in results.index.to_frame()["transform"].unique()]
    else:
        transforms = [None]

    figure_list: list[tuple[figure.Figure, plt.Axes]] = []
    for dataset in results.index.to_frame()["dataset"].unique():
        dataset_: str = str(dataset)
        for transform in transforms:
            for x_axis, y_axis in possible_pairs:
                fig: figure.Figure
                plot: plt.Axes
                fig, plot = plt.subplots(dpi=dpi)

                xtuple = (x_axis, x_axis.replace("_", " "))
                ytuple = (y_axis, y_axis.replace("_", " "))
                legend = single_plot(
                    plot, results, xtuple, ytuple, dataset_, transform, ptype=ptype
                )
                plt.tight_layout()

                if legend is False:  # use "is" here because otherwise any falsy value would match
                    plt.close(fig)
                    continue  # nothing was plotted -> don't save it and don't add it to the list

                if save:
                    fig.savefig(
                        directory / f"{dataset} {transform} {x_axis} {y_axis}.pdf",  # type: ignore
                        bbox_inches="tight",
                    )
                plt.close(fig)
                figure_list += [(fig, plot)]
    return figure_list
